#!/usr/bin/env python
# coding:utf-8
# Contributor:
#      fffonion        <fffonion@gmail.com>

import os
import re
import copy
import json
import uuid
import shutil
import zipfile
import tempfile
from threading import RLock
from . import util
from .const import *
from .const import __version__
if PY3K:
    from queue import Queue, Empty
else:
    from Queue import Queue, Empty

class Task(object):
    def __init__(self, url, cfgdict):
        self.url = url
        if url:
            _ = RE_INDEX.findall(url)
            if _:
                self.gid, self.sethash = _[0]
        self.failcode = 0
        self.state = TASK_STATE_WAITING
        self.guid = str(uuid.uuid4())[:8]
        self.config = cfgdict
        self.meta = {}
        self.has_ori = False
        self.reload_map = {} # {url:reload_url}
        self.filehash_map = {} # map same hash to different ids, {url:((id, fname), )}
        self.renamed_map = {} # map fid to renamed file name, used in finding a file by id in RPC
        self.img_q = None
        self.page_q = None
        self.list_q = None
        self._flist_done = set() # store id, don't save, will generate when scan
        self._monitor = None
        self._cnt_lock = RLock()
        self._f_lock = RLock()

    def cleanup(self, before_delete=False):
        if before_delete:
            if 'delete_task_files' in self.config and self.config['delete_task_files'] and \
                'title' in self.meta: # maybe it's a error task and meta is empty
                fpath = self.get_fpath()
                # TODO: ascii can't decode? locale not enus, also check save_file
                if os.path.exists(fpath):
                    shutil.rmtree(fpath)
                zippath = "%s.zip" % fpath
                if os.path.exists(zippath):
                    os.remove(zippath)
        elif self.state in (TASK_STATE_FINISHED, TASK_STATE_FAILED):
            self.img_q = None
            self.page_q = None
            self.list_q = None
            self.reload_map = {}
        
            # if 'filelist' in self.meta:
            #     del self.meta['filelist']
            # if 'resampled' in self.meta:
            #     del self.meta['resampled']

    def set_fail(self, code):
        self.state = TASK_STATE_FAILED
        self.failcode = code
        # cleanup all we cached
        self.meta = {}

    def migrate_exhentai(self):
        _ = re.findall("(?:https*://[g\.]*e\-hentai\.org)(.+)", self.url)
        if not _:
            return False
        self.url = "https://exhentai.org%s" % _[0]
        self.state = TASK_STATE_WAITING if self.state == TASK_STATE_FAILED else self.state
        self.failcode = 0
        return True

    def update_meta(self, meta):
        self.meta.update(meta)
        if self.config['jpn_title'] and self.meta['gjname']:
            self.meta['title'] = self.meta['gjname']
        else:
            self.meta['title'] = self.meta['gnname']

    # def guess_ori(self):
    #     # guess if this gallery has resampled files depending on some sample hashes
    #     # return True if it's ori
    #     if 'sample_hash' not in self.meta:
    #         return
    #     all_keys = map(lambda x:x[:10], self.meta['filelist'].keys())
    #     for h in self.meta['sample_hash']:
    #         if h not in all_keys:
    #             self.has_ori = True
    #             break
    #     del self.meta['sample_hash']

    def base_url(self):
        return re.findall(RESTR_SITE, self.url)[0]

    # def get_picpage_url(self, pichash):
    #     # if file resized, this url not works
    #     # http://%s.org/s/hash_s/gid-picid'
    #     return "%s/s/%s/%s-%s" % (
    #         self.base_url(), pichash[:10], self.gid, self.meta['filelist'][pichash][0]
    #     )

    def set_reload_url(self, imgurl, reload_url, fname):
        # if same file occurs severl times in a gallery
        if imgurl in self.reload_map:
            fpath = self.get_fpath()
            old_fid = self.get_fname(imgurl)[0]
            old_f = os.path.join(fpath, self.get_fidpad(old_fid))
            this_fid = int(RE_GALLERY.findall(reload_url)[0][1])
            this_f = os.path.join(fpath, self.get_fidpad(this_fid))
            self._f_lock.acquire()
            if os.path.exists(old_f):
                # we can just copy old file if already downloaded
                try:
                    with open(old_f, 'rb') as _of:
                        with open(this_f, 'wb') as _nf:
                            _nf.write(_of.read())
                except Exception as ex:
                    self._f_lock.release()
                    raise ex
                else:
                    self._f_lock.release()
                    self._cnt_lock.acquire()
                    self.meta['finished'] += 1
                    self._cnt_lock.release()
            else:
                # if not downloaded, we will copy them in save_file
                if imgurl not in self.filehash_map:
                    self.filehash_map[imgurl] = []
                self.filehash_map[imgurl].append((this_fid, old_fid))
                self._f_lock.release()
        else:
            self.reload_map[imgurl] = [reload_url, fname]

    def get_reload_url(self, imgurl):
        if not imgurl:
            return
        return self.reload_map[imgurl][0]

    def scan_downloaded(self, scaled = True):
        fpath = self.get_fpath()
        donefile = False
        if os.path.exists(os.path.join(fpath, ".xehdone")) or os.path.exists("%s.zip" % fpath):
            donefile = True
        _range_idx = 0
        for fid in range(1, self.meta['total'] + 1):
            # check download range
            if self.config['download_range']:
                _found = False
                # download_range is sorted asc
                for start, end in self.config['download_range'][_range_idx:]:
                    if fid > end: # out of range right bound move to next range
                        _range_idx += 1
                    elif start <= fid <= end: # in range
                        _found = True
                        break
                    elif fid < start: # out of range left bound
                        break
                if not _found:
                    self._flist_done.add(int(fid))
                    continue
            # can only check un-renamed files
            fname = os.path.join(fpath, self.get_fidpad(fid)) # id
            if donefile:
                self._flist_done.add(int(fid))
            elif os.path.exists(fname):
                if os.stat(fname).st_size == 0:
                    os.remove(fname)
                else:
                    self._flist_done.add(int(fid))
        self.meta['finished'] = len(self._flist_done)
        if self.meta['finished'] == self.meta['total']:
            self.state == TASK_STATE_FINISHED

    def queue_wrapper(self, callback, pichash = None, url = None):
        # if url is not finished, call callback to put into queue
        # type 1: normal file; type 2: resampled url
        # if pichash:
        #     fid = int(self.meta['filelist'][pichash][0])
        #     if fid not in self._flist_done:
        #         callback(self.get_picpage_url(pichash))
        # elif url:
        fhash, fid = RE_GALLERY.findall(url)[0]
        # if fhash not in self.meta['filelist']:
        #     self.meta['resampled'][fhash] = int(fid)
        #     self.has_ori = True]
        if int(fid) not in self._flist_done:
            callback(url)

    def save_file(self, imgurl, redirect_url, binary_iter):
        # TODO: Rlock for finished += 1
        fpath = self.get_fpath()
        self._f_lock.acquire()
        if not os.path.exists(fpath):
            os.mkdir(fpath)
        self._f_lock.release()
        pageurl, fname = self.reload_map[imgurl]
        _ = re.findall("/([^/\?]+)(?:\?|$)", redirect_url)
        if _: # change it if it's a full image
            fname = _[0]
            self.reload_map[imgurl][1] = fname
        _, fid = RE_GALLERY.findall(pageurl)[0]

        fn = os.path.join(fpath, self.get_fidpad(int(fid)))
        if os.path.exists(fn) and os.stat(fn).st_size > 0:
            return fn
        # create a femp file first
        # we don't need _f_lock because this will not be in a sequence
        # and we can't do that otherwise we are breaking the multi threading
        fd_tmp, fn_tmp = tempfile.mkstemp(prefix="xehentai-")
        os.close(fd_tmp)
        try:
            with open(fn_tmp, "wb") as f:
                for binary in binary_iter():
                    if self._monitor._exit(None):
                        raise DownloadAbortedException()
                    f.write(binary)
        except DownloadAbortedException as ex:
            try:
                os.unlink(fn_tmp)
            except:
                pass
            return

        self._f_lock.acquire()
        try:
            try:
                os.rename(fn_tmp, fn)
            except WindowsError as ex:
                # file is used by another process
                # do a copy and delete, WindowsError[32]
                if ex.errno == 13:
                    shutil.copy(fn_tmp, fn)
                    try:
                        os.unlink(fn_tmp)
                    except:
                        pass
                else:
                    raise ex
            self._cnt_lock.acquire()
            self.meta['finished'] += 1
            self._cnt_lock.release()
            if imgurl in self.filehash_map:
                for _fid, _ in self.filehash_map[imgurl]:
                    # if a file download is interrupted, it will appear in self.filehash_map as well
                    if _fid == int(fid):
                        continue
                    fn_rep = os.path.join(fpath, self.get_fidpad(_fid))
                    shutil.copyfile(fn, fn_rep)
                    self._cnt_lock.acquire()
                    self.meta['finished'] += 1
                    self._cnt_lock.release()
                del self.filehash_map[imgurl]
        except Exception as ex:
            self._f_lock.release()
            raise ex
        self._f_lock.release()
        return True

    def get_fname(self, imgurl):
        pageurl, fname = self.reload_map[imgurl]
        _, fid = RE_GALLERY.findall(pageurl)[0]
        return int(fid), fname

    def get_fpath(self):
        return os.path.join(self.config['dir'], util.legalpath(self.meta['title']))

    def get_fidpad(self, fid, ext = 'jpg'):
        fid = int(fid)
        _ = "%%0%dd.%%s" % (len(str(self.meta['total'])))
        return _ % (fid, ext)

    def rename_fname(self):
        fpath = self.get_fpath()
        tmppath = os.path.join(fpath, RENAME_TMPDIR)
        cnt = 0
        error_list = []
        # we need to track renamed fid's to decide 
        # whether to rename into a temp filename or add (1)
        # only need it when rename_ori = True
        done_list = set()
        for h in self.reload_map:
            fid, fname = self.get_fname(h)
            # if we don't need to rename to original name and file type matches
            if not self.config['rename_ori'] and os.path.splitext(fname)[1].lower() == '.jpg':
                continue
            fname_ori = os.path.join(fpath, self.get_fidpad(fid)) # id          
            if self.config['rename_ori']:
                if os.path.exists(os.path.join(tmppath, self.get_fidpad(fid))):
                    # if we previously put it into a temporary folder, we need to change fname_ori
                    fname_ori = os.path.join(tmppath, self.get_fidpad(fid))
                fname_to = os.path.join(fpath, util.legalpath(fname))
            else:
                # Q: Why we don't just use id.ext when saving files instead of using
                #   id.jpg?
                # A: If former task doesn't download all files, a new task with same gallery
                #   will have zero knowledge about file type before scanning all per page,
                #   thus can't determine if this id is downloaded, because file type is not
                #   necessarily .jpg
                fname_to = os.path.join(fpath, self.get_fidpad(fid, os.path.splitext(fname)[1][1:]))
            while fname_ori != fname_to:
                if os.path.exists(fname_ori):
                    while os.path.exists(fname_to):
                        _base, _ext = os.path.splitext(fname_to)
                        _ = re.findall("\((\d+)\)$", _base)
                        if self.config['rename_ori'] and fname_to not in done_list:
                            # if our auto numbering conflicts with original naming
                            # we move it into a temporary folder
                            # It's safe since this file is same with one of our auto numbering filename,
                            # it could never be conflicted with other files in tmppath
                            if not os.path.exists(tmppath):
                                os.mkdir(tmppath)
                            os.rename(fname_to, os.path.join(tmppath, os.path.split(fname_to)[1]))
                            break
                        if _ :# if ...(1) exists, use ...(2)
                            print(_base)
                            _base = re.sub("\((\d+)\)$", lambda x:"(%d)" % (int(x.group(1)) + 1), _base)
                        else:
                            _base = "%s(1)" % _base
                        fname_to = "".join((_base, _ext))
                try:
                    os.rename(fname_ori, fname_to)
                    self.renamed_map[str(fid)] = os.path.split(fname_to)[1]
                except Exception as ex:
                    error_list.append((os.path.split(fname_ori)[1], os.path.split(fname_to)[1], str(ex)))
                    break
                if self.config['rename_ori']:
                    done_list.add(fname_to)
                break
            cnt += 1
        if cnt == self.meta['total']:
            with open(os.path.join(fpath, ".xehdone"), "w"):
                pass
        try:
            os.rmdir(tmppath)
        except: # we will leave it undeleted if it's not empty
            pass
        return error_list

    def make_archive(self, remove=True):
        dpath = self.get_fpath()
        arc = "%s.zip" % dpath
        if os.path.exists(arc):
            return arc
        with zipfile.ZipFile(arc, 'w')  as zipFile:
            zipFile.comment = ("xeHentai Archiver v%s\nTitle:%s\nOriginal URL:%s" % (
                __version__, self.meta['title'], self.url)).encode('utf-8')
            for f in sorted(os.listdir(dpath)):
                fullpath = os.path.join(dpath, f)
                zipFile.write(fullpath, f, zipfile.ZIP_STORED)
        if remove:
            shutil.rmtree(dpath)
        return arc

    def from_dict(self, j):
        for k in self.__dict__:
            if k not in j:
                continue
            if k.endswith('_q') and j[k]:
                setattr(self, k, Queue())
                [getattr(self, k).put(e, False) for e in j[k]]
            else:
                setattr(self, k, j[k])
        _ = RE_INDEX.findall(self.url)
        if _:
            self.gid, self.sethash = _[0]
        return self


    def to_dict(self):
        d = dict({k:v for k, v in self.__dict__.items()
            if not k.endswith('_q') and not k.startswith("_")})
        for k in ['img_q', 'page_q', 'list_q']:
            if getattr(self, k):
                d[k] = [e for e in getattr(self, k).queue]
        return d



#!/usr/bin/env python
# coding:utf-8
# Contributor:
#      fffonion        <fffonion@gmail.com>

from __future__ import absolute_import
import os
import re
import sys
import math
import json
import time
import traceback
from .task import Task
from . import util
from . import proxy
from . import filters
from .rpc import RPCServer
from .i18n import i18n
from .util import logger
from .const import *
from .const import __version__
from .worker import *
if PY3K:
    from queue import Queue, Empty
else:
    from Queue import Queue, Empty

from . import config as default_config
sys.path.insert(1, FILEPATH)
try:
    import config
except ImportError:
    config = default_config
sys.path.pop(1)

class xeHentai(object):
    def __init__(self):
        self.verstr = "%.3f%s" % (__version__, '-dev' if DEVELOPMENT else "")
        self.logger = logger.Logger()
        self._exit = False
        self.tasks = Queue() # for queueing, stores gid only
        self.last_task_guid = None
        self._all_tasks = {} # for saving states
        self._all_threads = [[] for i in range(20)]
        self.cfg = {k:v for k,v in default_config.__dict__.items() if not k.startswith("_")}
        # note that ignored_errors are overwritten using val from custom config
        self.cfg.update({k:v for k,v in config.__dict__.items() if not k.startswith("_")})
        self.proxy = None
        self.cookies = {"nw": "1"}
        self.headers = {
            'User-Agent': util.make_ua(),
            'Accept-Charset': 'utf-8;q=0.7,*;q=0.7',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'Connection': 'keep-alive'
        }
        self.has_login = False
        self.load_session()
        self.rpc = None

    def update_config(self, **cfg_dict):
        self.cfg.update({k:v for k, v in cfg_dict.items() if k in cfg_dict and k not in ('ignored_errors',)})
        # merge ignored errors list
        if 'ignored_errors' in cfg_dict and cfg_dict['ignored_errors']:
            self.cfg['ignored_errors'] = list(set(self.cfg['ignored_errors'] + cfg_dict['ignored_errors']))
        self.logger.set_level(logger.Logger.WARNING - self.cfg['log_verbose'])
        self.logger.verbose("cfg %s" % self.cfg)
        if cfg_dict['proxy']:
            if not self.proxy: # else we keep it None
                self.proxy = proxy.Pool()
            for p in self.cfg['proxy']:
                try:
                    self.proxy.add_proxy(p)
                except Exception as ex:
                    self.logger.warning(traceback.format_exc())
            self.logger.debug(i18n.PROXY_CANDIDATE_CNT % len(self.proxy.proxies))
        if cfg_dict['dir'] and not os.path.exists(cfg_dict['dir']):
            try:
                os.makedirs(cfg_dict['dir'])
            except OSError as ex:  # Python >2.5
                self.logger.error(i18n.ERR_CANNOT_CREATE_DIR % cfg_dict['dir'])
        if not self.rpc and self.cfg['rpc_port'] and self.cfg['rpc_interface']:
            self.rpc = RPCServer(self, (self.cfg['rpc_interface'], int(self.cfg['rpc_port'])),
                secret = None if 'rpc_secret' not in self.cfg else self.cfg['rpc_secret'],
                logger = self.logger)
            if not RE_LOCAL_ADDR.match(self.cfg['rpc_interface']) and \
                not self.cfg['rpc_secret']:
                self.logger.warning(i18n.RPC_TOO_OPEN % self.cfg['rpc_interface'])
            self.rpc.start()
        self.logger.set_logfile(self.cfg['log_path'])
        return ERR_NO_ERROR, ""

    def _get_httpreq(self, proxy_policy):
        return HttpReq(self.headers, logger = self.logger, proxy = self.proxy, proxy_policy = proxy_policy)

    def _get_httpworker(self, tid, task_q, flt, suc, fail, keep_alive, proxy_policy, timeout, stream_mode):
        return HttpWorker(tid, task_q, flt, suc, fail,
            headers = self.headers, proxy = self.proxy, logger = self.logger,
            keep_alive = keep_alive, proxy_policy = proxy_policy, timeout = timeout, stream_mode = stream_mode)

    def add_task(self, url, **cfg_dict):
        url = url.strip()
        cfg = {k:v for k, v in self.cfg.items() if k in (
            "dir", "download_ori", "download_thread_cnt", "scan_thread_cnt",
            "proxy_image", "proxy_image_only", "ignored_errors",
            "rename_ori", "make_archive", "delete_task_files", "jpn_title", "download_range", "download_timeout")}
        cfg.update(cfg_dict)
        if cfg['download_ori'] and not self.has_login:
            self.logger.warning(i18n.XEH_DOWNLOAD_ORI_NEED_LOGIN)
        t = Task(url, cfg)
        if t.guid in self._all_tasks:
            if self._all_tasks[t.guid].state in (TASK_STATE_FINISHED, TASK_STATE_FAILED):
                self.logger.debug(i18n.TASK_PUT_INTO_WAIT % t.guid)
                self._all_tasks[t.guid].state = TASK_STATE_WAITING
                self._all_tasks[t.guid].cleanup()
            return 0, t.guid
        self._all_tasks[t.guid] = t
        if not re.match("^%s/[^/]+/\d+/[^/]+/*#*$" % RESTR_SITE, url):
            t.set_fail(ERR_URL_NOT_RECOGNIZED)
        elif not self.has_login and re.match("^https*://exhentai\.org", url):
            t.set_fail(ERR_CANT_DOWNLOAD_EXH)
        else:
            self.tasks.put(t.guid)
            return 0, t.guid
        self.logger.error(i18n.TASK_ERROR % (t.guid, i18n.c(t.failcode)))
        return t.failcode, None

    def del_task(self, guid):
        if guid not in self._all_tasks:
            return ERR_TASK_NOT_FOUND, None
        if TASK_STATE_PAUSED< self._all_tasks[guid].state < TASK_STATE_FINISHED:
            return ERR_DELETE_RUNNING_TASK, None
        self._all_tasks[guid].cleanup(before_delete=True)
        del self._all_tasks[guid]
        return ERR_NO_ERROR, ""

    def pause_task(self, guid):
        if guid not in self._all_tasks:
            return ERR_TASK_NOT_FOUND, None
        t = self._all_tasks[guid]
        if t.state in (TASK_STATE_PAUSED, TASK_STATE_FINISHED, TASK_STATE_FAILED):
            return ERR_TASK_CANNOT_PAUSE, None
        if t._monitor:
            t._monitor._exit = lambda x: True
        t.state = TASK_STATE_PAUSED
        return ERR_NO_ERROR, ""

    def resume_task(self, guid):
        if guid not in self._all_tasks:
            return ERR_TASK_NOT_FOUND, None
        t = self._all_tasks[guid]
        if TASK_STATE_PAUSED< t.state < TASK_STATE_FINISHED:
            return ERR_TASK_CANNOT_RESUME, None
        t.state = max(t.state, TASK_STATE_WAITING)

        self.tasks.put(guid)
        return ERR_NO_ERROR, ""

    def _do_task(self, task_guid):
        task = self._all_tasks[task_guid]
        if task.state == TASK_STATE_WAITING:
            task.state = TASK_STATE_GET_META
        req = self._get_httpreq(util.get_proxy_policy(task.config))
        if not task.page_q:
            task.page_q = Queue() # per image page queue
        if not task.img_q:
            task.img_q = Queue() # (image url, savepath) queue
        monitor_started = False
        while self._exit < XEH_STATE_FULL_EXIT:
            # wait for threads from former task to stop
            if self._all_threads[task.state]:
                self.logger.verbose("wait %d threads in state %s" % (
                    len(self._all_threads[task.state]), task.state))
                for t in self._all_threads[task.state]:
                    t.join()
                self._all_threads[task.state] = []
                # check again before we bring up new threads
                continue
            if task.state >= TASK_STATE_SCAN_IMG and not monitor_started:
                self.logger.verbose("state %d >= %d, bring up montior" % (task.state, TASK_STATE_SCAN_IMG))
                # bring up the monitor here, ahead of workers
                mon = Monitor(req, self.proxy, self.logger, task, ignored_errors=task.config['ignored_errors'])
                _ = ['down-%d' % (i + 1) for i in range(task.config['download_thread_cnt'])]
                # if we jumpstart from a saved session to DOQNLOAD
                # there will be no scan_thread
                # if task.state >= TASK_STATE_SCAN_PAGE:
                #    _ += ['list-1']
                if task.state >= TASK_STATE_SCAN_IMG:
                    _ += ['scan-%d' % (i + 1) for i in range(task.config['scan_thread_cnt'])]
                mon.set_vote_ns(_)
                self._monitor = mon
                task._monitor = mon
                mon.start()
                # put in the lowest state
                self._all_threads[TASK_STATE_SCAN_IMG].append(mon)
                monitor_started = True

            if task.state == TASK_STATE_GET_META: # grab meta data
                try:
                    r = req.request("GET", task.url,
                        filters.flt_metadata,
                        lambda x:task.update_meta(x),
                        lambda x:task.set_fail(x))
                except Exception as ex:
                    self.logger.error(i18n.TASK_ERROR % (task.guid, traceback.format_exc()))
                    task.state = TASK_STATE_FAILED
                    break
                if task.failcode in (ERR_ONLY_VISIBLE_EXH, ERR_GALLERY_REMOVED) and self.has_login and \
                        task.migrate_exhentai():
                    self.logger.info(i18n.TASK_MIGRATE_EXH % task_guid)
                    self.tasks.put(task_guid)
                    break
                elif task.failcode == ERR_IP_BANNED:
                    self.logger.error(i18n.c(ERR_IP_BANNED) % r)
                    task.state = TASK_STATE_FAILED
                    break

            # elif task.state == TASK_STATE_GET_HATHDL: # download hathdl
            #     r = req.request("GET",
            #         "%s/hathdler.php?gid=%s&t=%s" % (task.base_url(), task.gid, task.sethash),
            #         filters.flt_hathdl,
            #         lambda x:(task.meta.update(x),
            #             task.guess_ori(),
            #             task.scan_downloaded()),
            #                 #task.meta['has_ori'] and task.config['download_ori'])),
            #         lambda x:task.set_fail(x),)
            #     self.logger.info(i18n.TASK_WILL_DOWNLOAD_CNT % (
            #         task_guid, task.meta['total'] - len(task._flist_done),
            #         task.meta['total']))
            elif task.state == TASK_STATE_SCAN_PAGE:
                # if task.config['fast_scan'] and not task.has_ori:
                #     self.logger.info(i18n.TASK_FAST_SCAN % task.guid)
                #     for p in task.meta['filelist']:
                #         task.queue_wrapper(task.page_q.put, pichash = p)
                # else:
                # scan by our own, should not be here currently
                # start backup thread
                task.scan_downloaded()
                if task.state == TASK_STATE_FINISHED:
                    continue
                for x in range(0,
                    int(math.ceil(1.0 * task.meta['total'] / int(task.meta['thumbnail_cnt'])))):
                    r = req.request("GET",
                        "%s/?p=%d" % (task.url, x),
                        filters.flt_pageurl,
                        lambda x: task.queue_wrapper(task.page_q.put, url = x),
                        lambda x: task.set_fail(x))
                    if task.failcode:
                        break
            elif task.state == TASK_STATE_SCAN_IMG:
                # print here so that see it after we can join former threads
                self.logger.info(i18n.TASK_TITLE % (
                    task_guid, task.meta['title']))
                self.logger.info(i18n.TASK_WILL_DOWNLOAD_CNT % (
                    task_guid, task.meta['total'] - task.meta['finished'],
                    task.meta['total']))
                # spawn thread to scan images
                for i in range(task.config['scan_thread_cnt']):
                    tid = 'scan-%d' % (i + 1)
                    _ = self._get_httpworker(tid, task.page_q,
                        filters.flt_imgurl_wrapper(task.config['download_ori'] and self.has_login),
                        lambda x, tid = tid: (task.set_reload_url(x[0], x[1], x[2]),
                            task.img_q.put(x[0]),
                            mon.vote(tid, 0)),
                        lambda x, tid = tid: (mon.vote(tid, x[0])),
                        mon.wrk_keepalive,
                        util.get_proxy_policy(task.config),
                        10,
                        False)
                        # we don't need proxy_image in the scan thread
                        # we use default timeout in the scan thread
                    # _._exit = lambda t: t._finish_queue()
                    self._all_threads[TASK_STATE_SCAN_IMG].append(_)
                    _.start()
                task.state = TASK_STATE_DOWNLOAD - 1
            elif task.state == TASK_STATE_SCAN_ARCHIVE:
                task.state = TASK_STATE_DOWNLOAD - 1
            elif task.state == TASK_STATE_DOWNLOAD:
                # spawn thread to download all urls
                for i in range(task.config['download_thread_cnt']):
                    tid = 'down-%d' % (i + 1)
                    _ = self._get_httpworker(tid, task.img_q,
                        filters.download_file_wrapper(task.config['dir']),
                        lambda x, tid = tid: (task.save_file(x[1], x[2], x[0]) and \
                            (self.logger.debug(i18n.XEH_FILE_DOWNLOADED.format(tid, *task.get_fname(x[1]))),
                                mon.vote(tid, 0))),
                        lambda x, tid = tid: (
                            task.page_q.put(task.get_reload_url(x[1])),# if x[0] != ERR_QUOTA_EXCEEDED else None,
                            task.reload_map.pop(x[1]) if x[1] in task.reload_map else None, # delete old url in reload_map
                            self.logger.debug(i18n.XEH_DOWNLOAD_HAS_ERROR % (tid, i18n.c(x[0]))),
                            mon.vote(tid, x[0])),
                        mon.wrk_keepalive,
                        util.get_proxy_policy(task.config),
                        task.config['download_timeout'],
                        True)
                    self._all_threads[TASK_STATE_DOWNLOAD].append(_)
                    _.start()
                # spawn archiver if we need
                if task.config['make_archive']:
                    if self._all_threads[TASK_STATE_MAKE_ARCHIVE]:
                        self._all_threads[TASK_STATE_MAKE_ARCHIVE][0].join()
                        self._all_threads[TASK_STATE_MAKE_ARCHIVE] = []
                    _a = ArchiveWorker(self.logger, task)
                    self._all_threads[TASK_STATE_MAKE_ARCHIVE].append(_a)
                    _a.start()
                # break current task loop
                break

            if task.failcode:
                self.logger.error(i18n.TASK_ERROR % (task_guid, i18n.c(task.failcode)))
                # wait all threads to finish
                break
            else:
                task.state += 1

    def _task_loop(self):
        task_guid = None
        cnt = 0
        while not self._exit:
            # get a new task
            if cnt == 10:
                self.save_session()
                cnt = 0
            try:
                _ = self.tasks.get(False)
                self.last_task_guid = task_guid
                task_guid = _
            except Empty:
                time.sleep(1)
                cnt += 1
                continue
            else:
                task = self._all_tasks[task_guid]
                if TASK_STATE_PAUSED < task.state < TASK_STATE_FINISHED:
                    self.logger.info(i18n.TASK_START % task_guid)
                    self.save_session()
                    cnt = 0
                    self._do_task(task_guid)
        self.logger.info(i18n.XEH_LOOP_FINISHED)
        self._cleanup()

    def _term_threads(self):
        self._exit = XEH_STATE_FULL_EXIT
        for l in self._all_threads:
            for p in l:
                p._exit = lambda x:True

    def _cleanup(self):
        self._exit = self._exit if self._exit > 0 else XEH_STATE_SOFT_EXIT
        self.save_session()
        self._join_all()
        self.logger.cleanup()
        # let's send a request to rpc server to unblock it
        if self.rpc:
            self.rpc._exit = lambda x:True
            import requests
            try:
                requests.get("http://%s:%s/" % (self.cfg['rpc_interface'], self.cfg['rpc_port']))
            except:
                pass
            self.rpc.join()
        # save it again in case we miss something
        self.save_session()
        self._exit = XEH_STATE_CLEAN

    def _join_all(self):
        for l in self._all_threads:
            for p in l:
                p.join()

    def save_session(self):
        with open("h.json", "w") as f:
            try:
                f.write(json.dumps({
                    'tasks':{} if not self.cfg['save_tasks'] else
                        {k: v.to_dict() for k,v in self._all_tasks.items()},
                    'cookies':self.cookies}))
            except Exception as ex:
                self.logger.warning(i18n.SESSION_WRITE_EXCEPTION % traceback.format_exc())
                return ERR_SAVE_SESSION_FAILED, str(ex)
        return ERR_NO_ERROR, None

    def load_session(self):
        if os.path.exists("h.json"):
            with open("h.json") as f:
                try:
                    j = json.loads(f.read())
                except Exception as ex:
                    self.logger.warning(i18n.SESSION_LOAD_EXCEPTION % traceback.format_exc())
                    return ERR_SAVE_SESSION_FAILED, str(ex)
                else:
                    for _ in j['tasks'].values():
                        _t = Task("", {}).from_dict(_)
                        if 'filelist' in _t.meta:
                            _t.scan_downloaded()
                                #_t.meta['has_ori'] and task.config['download_ori'])
                        # since we don't block on scan_img state, an unempty page_q
                        # indicates we should start from scan_img state,
                        if _t.state == TASK_STATE_DOWNLOAD and _t.page_q:
                            _t.state = TASK_STATE_SCAN_IMG
                        self._all_tasks[_['guid']] = _t
                        self.tasks.put(_['guid'])
                    if self._all_tasks:
                        self.logger.info(i18n.XEH_LOAD_TASKS_CNT % len(self._all_tasks))
                    self.cookies.update(j['cookies'])
                    if self.cookies:
                        self.headers.update({'Cookie':util.make_cookie(self.cookies)})
                        self.has_login = 'ipb_member_id' in self.cookies and 'ipb_pass_hash' in self.cookies
        _1xcookie = os.path.join(FILEPATH, ".ehentai.cookie")# 1.x cookie file
        if not self.has_login and os.path.exists(_1xcookie):
            with open(_1xcookie) as f:
                try:
                    cid, cpw = f.read().strip().split(",")
                    self.cookies.update({'ipb_member_id':cid, 'ipb_pass_hash':cpw})
                    self.headers.update({'Cookie':util.make_cookie(self.cookies)})
                    self.has_login = True
                    self.logger.info(i18n.XEH_LOAD_OLD_COOKIE)
                except:
                    pass

        return ERR_NO_ERROR, None

    def login_exhentai(self, name, pwd):
        if 'ipb_member_id' in self.cookies and 'ipb_pass_hash' in self.cookies:
            return
        self.logger.debug(i18n.XEH_LOGIN_EXHENTAI)
        logindata = {
            'UserName':name,
            'returntype':'8',
            'CookieDate':'1',
            'b':'d',
            'bt':'pone',
            'PassWord':pwd
        }
        req = self._get_httpreq(util.get_proxy_policy(self.cfg))
        req.request("POST", "https://forums.e-hentai.org/index.php?act=Login&CODE=01",
            filters.login_exhentai,
            lambda x:(
                setattr(self, 'cookies', x),
                setattr(self, 'has_login', True),
                self.headers.update({'Cookie':util.make_cookie(self.cookies)}),
                self.save_session(),
                self.logger.info(i18n.XEH_LOGIN_OK)),
            lambda x:(self.logger.warning(str(x)),
                self.logger.info(i18n.XEH_LOGIN_FAILED)),
            logindata)
        return ERR_NO_ERROR, self.has_login

    def set_cookie(self, cookie):
        self.cookies.update(util.parse_cookie(cookie))
        self.headers.update({'Cookie':util.make_cookie(self.cookies)})
        if 'ipb_member_id' in self.cookies and 'ipb_pass_hash' in self.cookies:
            self.has_login = True
        return ERR_NO_ERROR, None


if __name__ == '__main__':
    pass



#!/usr/bin/env python
# coding:utf-8
# Contributor:
#      fffonion        <fffonion@gmail.com>

import re
import time
import random
from requests.exceptions import ConnectTimeout, ConnectionError, ProxyError, InvalidSchema
from requests.packages.urllib3.exceptions import ProxySchemeUnknown
from . import util
from .const import *

MAX_FAIL = 5

class PoolException(Exception):
    pass

class Pool(object):
    def __init__(self, disable_policy = None):
        self.proxies = {}
        self.errors = {}
        if not disable_policy:
            self.disable_policy = lambda x, y: y >= MAX_FAIL
        else:
            self.disable_policy = disable_policy
        self.disabled = {} # key: expire

    def proxied_request(self, session):
        for d in self.disabled:
            if 0 < self.disabled[d] < time.time():
                try:
                    del self.disabled[d]
                except:
                    pass
        l = [i for i in self.proxies.keys() if i not in self.disabled]
        if not l:
            raise PoolException("try to use proxy but no proxies avaliable")
        # _ = self.proxies[random.choice(l)]
        _ = self.proxies[l[0]]
        return _[0](session), self.not_good(l[0])

    def has_available_proxies(self):
        return len([i for i in self.proxies.keys() if i not in self.disabled]) == 0

    def not_good(self, addr):
        def n(weight = MAX_FAIL, expire = 0):
            self.proxies[addr][2] += weight
            if self.disable_policy(*self.proxies[addr][1:]):
                # add to disabled set
                self.disabled[addr] = expire + time.time()
        return n

    def trace_proxy(self, addr, weight = 1, check_func = None, exceptions = []):
        def _(func):
            def __(*args, **kwargs):
                ex = None
                try:
                    r = func(*args, **kwargs)
                except Exception as _ex:
                    ex = _ex
                    for e in [ConnectTimeout, ConnectionError, ProxyError] + exceptions:
                        if isinstance(ex, e):
                            # ignore BadStatusLine, this doesn't mean the proxy is bad
                            if e == ConnectionError and 'BadStatusLine' in str(e):
                                continue
                            self.proxies[addr][2] += weight
                            break
                else:
                    if check_func and not check_func(r):
                        self.proxies[addr][2] += weight
                    else:
                        # suc count + 1
                        self.proxies[addr][1] += weight
                if self.disable_policy(*self.proxies[addr][1:]):
                    # add to disabled set and never expire
                    self.disabled[addr] = 0
                # print(self.proxies[addr])
                if ex:
                    # import traceback
                    # traceback.print_exc()
                    raise ex
                return r
            return __
        return _

    def add_proxy(self, addr):
        if re.match("socks[45][ah]*://([^:^/]+)(\:\d{1,5})*/*$", addr):
            p = socks_proxy(addr, self.trace_proxy)
        elif re.match("https*://([^:^/]+)(\:\d{1,5})*/*$", addr):
            p = http_proxy(addr, self.trace_proxy)
        elif re.match("https*://([^:^/]+)(\:\d{1,5})*/.+\.php\?.*b=.+", addr):
            p = glype_proxy(addr, self.trace_proxy)
        else:
            raise ValueError("%s is not an acceptable proxy address" % addr)
        self.proxies[addr] = [p, 0, 0]

def socks_proxy(addr, trace_proxy):
    proxy_info = {
        'http':addr,
        'https':addr
    }
    def handle(session):
        @trace_proxy(addr, exceptions = [ProxySchemeUnknown, InvalidSchema])
        def f(*args, **kwargs):
            kwargs.update({'proxies': proxy_info})
            return session.request(*args, **kwargs)
        return f
    return handle

def http_proxy(addr, trace_proxy):
    proxy_info = {
        'http':addr,
        'https':addr
    }
    def handle(session):
        @trace_proxy(addr)
        def f(*args, **kwargs):
            kwargs.update({'proxies': proxy_info})
            return session.request(*args, **kwargs)
        return f
    return handle

def glype_proxy(addr, trace_proxy):
    g_session = {"s":""}
    def handle(session, g_session = g_session):
        import urllib
        argname = re.findall('[&\?]([a-zA-Z\._]+)=[^\d]*', addr)[0]
        bval = re.findall('[&\?]b=(\d*)', addr)
        bval = bval[0] if bval else '4'
        server, inst_loc, script = re.findall('(https*://[^/]+)/(.*?)([^/]+\.php)', addr)[0]
        urlre = re.compile('/%s%s\?u=([^&"\']+)&[^"\']+' % (inst_loc, script))
        def mkurl(url):
            return "%s/%s%s?%s=%s&b=%s&f=norefer" % (
                server, inst_loc, script, argname,
                (urllib.parse if PY3K else urllib).quote_plus(url), bval)
        @trace_proxy(addr)
        def f(*args, **kwargs):
            # change url
            url = args[1]
            args = (args[0], mkurl(url),)
            kwargs['headers'] = dict(session.headers)
            # anti hotlinking
            kwargs['headers'].update({'Referer':"%s/%s%s" % (server, inst_loc, script)})
            _coo_new = dict(g_session) if g_session['s'] else {}
            if 'Cookie' in kwargs['headers']:
                site = re.findall('https*://([^/]+)/*', url)[0]
                _coo_old = util.parse_cookie(kwargs['headers']['Cookie'])
                for k in _coo_old:
                    _coo_new["c[%s][/][%s]" % (site, k)] = _coo_old[k]
                kwargs['headers']['Cookie'] = util.make_cookie(_coo_new)
            tried = 0
            while True:
                if tried == 2:
                    raise PoolException("can't bypass glype https warning")
                rt = session.request(*args, **kwargs)
                if '<input type="hidden" name="action" value="sslagree">' not in rt.text:
                    break
                rt = session.request("GET", "%s/%sincludes/process.php?action=sslagree" % (server, inst_loc),
                    allow_redirects = False, **kwargs)
                if rt.headers.get('set-cookie'):
                    _coo_new.update(util.parse_cookie(rt.headers.get('set-cookie').replace(",", ";")))
                    kwargs['headers']['Cookie'] = util.make_cookie(_coo_new)
                    if 's' in _coo_new:
                        g_session["s"] = _coo_new['s']
                        # print(g_session)
                tried += 1

            if rt.headers.get('set-cookie'):
                coo = util.parse_cookie(rt.headers.get('set-cookie').replace(",", ";"))
                for k in list(coo.keys()):
                    _ = re.findall('c\[[^]]+\]\[[^]]+\]\[([^]]+)\]', k)
                    if _:
                        coo[_[0]] = coo[k]
                rt.headers['set-cookie'] = util.make_cookie(coo)
            # change url back, only change on text/* mime types
            rt.url = url
            if rt.headers.get('content-type').startswith("text"):
                if PY3K:
                    rt._content = rt._content.decode('utf-8')
                _ = re.match('<div id="error">(.*?)</div>', rt.content)
                if _:
                    raise PoolException("glype returns: %s" % _[0])
                # change transformed url back
                rt._content = urlre.sub(lambda x:(urllib.parse if PY3K else urllib).unquote(x.group(1)), rt._content)
                if PY3K:
                    rt._content = rt._content.encode('utf-8')
            return rt

        return f
    return handle

if __name__ == '__main__':
    import requests
    p = Pool()
    p.add_proxy("sock5://127.0.0.1:16961")
    print(p.proxied_request(requests.Session())("GET", "http://ipip.tk", headers = {}, timeout = 2).headers)



import requests
import random
try:
    from urlparse import urlparse, urlunparse
except:
    from urllib.parse import urlparse, urlunparse

class FallbackIpAdapter(requests.adapters.HTTPAdapter):
    def __init__(self, ip_map, **kwargs):
        self.ip_map = ip_map
        requests.adapters.HTTPAdapter.__init__(self, **kwargs)

    # override
    def get_connection(self, url, proxies=None):
        if not proxies:
            parsed = urlparse(url)
            _hostname = parsed.hostname
            _scheme = parsed.scheme
            if _hostname in self.ip_map:
                _parsed = list(parsed)
                # alter the hostname
                _hostname = '%s%s' % (random.choice(self.ip_map[_hostname]),
                                        (":%d" % parsed.port) if parsed.port else "")
                _scheme = 'https'
                print(parsed.scheme, _scheme)
            return self.poolmanager.connection_from_host(_hostname, parsed.port, scheme=_scheme,
                                                            pool_kwargs={'assert_hostname': parsed.hostname})
        else:
            # fallback
            return requests.adapters.HTTPAdapter.get_connection(self, url, proxies)
    
    def add_headers(self, request, **kwargs):
        if not request.headers.get('Host'):
            parsed = urlparse(request.url)
            request.headers['Host'] = parsed.hostname

    def cert_verify(self, conn, url, verify, cert):
        if url.startswith('http://'):
            url = "https://%s" % url[7:]
        return requests.adapters.HTTPAdapter.cert_verify(self, conn, url, verify, cert)
    
    # def __getattr__(self, k):
        # fallback attribute handler
    #    return getattr(self.super, k)
_FALLBACK_CF_IP = ("104.24.255.11", "104.24.254.11")

s = requests.Session()
s.mount('https://', FallbackIpAdapter({
    'e-hentai.org': _FALLBACK_CF_IP,
    'forums.e-hentai.org': ("94.100.18.243", ) + _FALLBACK_CF_IP,
    'exhentai.org': ("217.23.13.91","217.23.13.45","109.236.84.136","109.236.92.143","109.236.84.145","109.236.92.166")
} ))
s.mount('http://', FallbackIpAdapter({
    'e-hentai.org': _FALLBACK_CF_IP,
    'forums.e-hentai.org': ("94.100.18.243", ) + _FALLBACK_CF_IP,
    'exhentai.org': ("217.23.13.91","217.23.13.45","109.236.84.136","109.236.92.143","109.236.84.145","109.236.92.166")
} ))
# print(s.get('http://httpbin.tk/headers', proxies = {'http':'socks5://127.0.0.1:16961'}).content)
print(s.get('https://exhentai.org', allow_redirects=False).headers)


#!/usr/bin/env python
# coding:utf-8
# Contributor:
#      fffonion        <fffonion@gmail.com>

import os
import re
from . import util
from .const import *

SUC = 0
FAIL = 1

def login_exhentai(r, suc, fail):
    # input login response
    # add cookies if suc; log error fail
    try:
        coo = r.headers.get('set-cookie')
        cooid = re.findall('ipb_member_id=(.*?);', coo)[0]
        coopw = re.findall('ipb_pass_hash=(.*?);', coo)[0]
    except (IndexError, ) as ex:
        errmsg = re.findall('<span class="postcolor">([^<]+)</span>', r.text)
        if errmsg:
            fail(errmsg[0])
        else:
            fail("ex: %s" % ex)
        return FAIL
    else:
        suc({'ipb_member_id':cooid, 'ipb_pass_hash':coopw})
        return SUC


def flt_metadata(r, suc, fail):
    # input index response
    # add gallery meta if suc; return errorcode if fail
    # TODO: catch re exceptions
    if r.status_code == 404:
        return fail(ERR_GALLERY_REMOVED)
    if re.match("This gallery is pining for the fjords", r.text):
        return fail(ERR_ONLY_VISIBLE_EXH)
    elif re.match("Your IP address has been temporarily banned", r.text):
        fail(ERR_IP_BANNED)
        return re.findall("The ban expires in (.+)", r.text)[0]
    meta = {}
    # print(r.text)
    # sample_hash = re.findall('<a href="%s/./([a-f0-9]{10})/\d+\-\d+"><img' % RESTR_SITE, r.text)
    # meta['sample_hash'] = sample_hash
    # meta['resampled'] = {}
    meta['gjname'] = util.htmlescape(re.findall('="gj">(.*?)</h1>', r.text)[0])
    meta['gnname']= util.htmlescape(re.findall('="gn">(.*?)</h1>', r.text)[0])
    # don't assign title now, select gj/gn based on cfg['jpn_title']
    # meta['title'] = meta['gjname'] if meta['gjname'] else meta['gnname']
    meta['total'] = int(re.findall('Length:</td><td class="gdt2">(\d+)\s+page', r.text)[0])
    meta['finished'] = 0
    meta['tags'] = re.findall("toggle_tagmenu\('([^']+)'", r.text)

    # TODO: parse cookie to calc thumbnail_cnt (tr_2, ts_m)
    _ = re.findall("Showing (\d+) \- (\d+) of ([\d,]+) images", r.text)[0]
    meta['thumbnail_cnt'] = int(_[1]) - int(_[0]) + 1

    suc(meta)
    # _ = re.findall(
    #    '%s/[^/]+/(\d+)/[^/]+/\?p=\d*" onclick="return false"(.*?)</a>' % RESTR_SITE,
    #    r.text)
    # meta['pagecount'] = 1 if len(_) <= 1 else int(pagecount[-2])


# def flt_hathdl(r, suc, fail):
#     # input hathdl response
#     # add gallery meta if suc; return errorcode if fail
#     if r.status_code == 404:
#         fail(ERR_HATHDL_NOTFOUND)
#         return ERR_HATHDL_NOTFOUND
#     try:
#         meta = {
#             'name': util.htmlescape(re.findall('TITLE (.+)', r.text)[0]),
#             #'gid': int(re.findall('GID (.+)', r.text)[0]),
#             'total': int(re.findall('FILES (.+)', r.text)[0]),
#             'finished': 0,
#             'title': re.findall('Title:\s+(.+)', r.text)[0],
#             #'upload_time': re.findall('Upload Time:\s+(.+)', r.text)[0], # invisible
#             #'upload_by': re.findall('Uploaded By:\s+(.+)', r.text)[0], # invisible
#             #'downloaded': re.findall('Downloaded:\s+(.+)', r.text)[0], # invisible
#             'tags': re.findall('Tags:\s+(.+)', r.text)[0].split(', '),
#         }
#         listtmp = re.findall('FILELIST\n(.+)\n+\nINFORMATION', r.text, re.DOTALL)[0].split('\n')
#         meta['filelist'] = {}
#         for l in listtmp:
#             # hash(full): id, hash_10, length, width, height, format, name
#             _ = re.findall('(\d+) ([a-z0-9]+)-(\d+)-(\d+)-(\d+)-([a-z]+) (.+)', l)[0]
#             meta['filelist'][_[1][:10]] = list(_)
#     except (IndexError, ValueError) as ex:
#         fail(ERR_MALFORMED_HATHDL)
#         return ERR_MALFORMED_HATHDL
#     suc(meta)

def flt_pageurl(r, suc, fail):
    # input gallery response
    # add per image urls if suc; finish task if fail
    picpage = re.findall(
        '<a href="(%s/./[a-f0-9]{10}/\d+\-\d+)"><img alt="\d+" title="Page' % RESTR_SITE,
        r.text)
    if not picpage:
        fail(ERR_NO_PAGEURL_FOUND)
    for p in picpage:
        suc(p)

def flt_quota_check(func):
    def _(r, suc, fail):
        if r.status_code == 600:# tcp layer error
            fail((ERR_CONNECTION_ERROR, r._real_url))
        elif r.status_code == 403:
            fail((ERR_KEY_EXPIRED, r._real_url))
        elif r.status_code == 509 or r.content_length in [925, 28658, 144, 210, 1009] or '509.gif' in r.url:
            fail((ERR_QUOTA_EXCEEDED, r._real_url))
            # will not call the decorated filter
        elif r.content_length < 200 and \
                r.headers.get('content-type') and r.headers.get('content-type').startswith('text') and \
                re.findall("exceeded your image viewing limits", r.text):
            fail((ERR_QUOTA_EXCEEDED, r._real_url))
            # will not call the decorated filter
        else:
            func(r, suc, fail)
    return _

def flt_imgurl_wrapper(ori):
    @flt_quota_check
    def flt_imgurl(r, suc, fail, ori = ori):
        # input per image page response
        # add (image url, reload url, filename) to queue if suc
        # return (errorcode, page_url) if fail
        if re.match('Invalid page', r.text):
            return fail((ERR_IMAGE_RESAMPLED, r._real_url))
        while True:
            _ = re.findall('src="([^"]+keystamp[^"]+)"', r.text)
            if not _:
                _ = re.findall('src="([^"]+)"\s+style="', r.text)
            if not _:
                break
            picurl = util.htmlescape(_[0])

            _ = re.findall('</a></div><div>(.*?) ::.+::.+</di', r.text)
            if not _:
                break
            filename = _[0]
            if 'image.php' in filename:
                _ = re.findall('n=(.+)', picurl)
                if not _:
                    break
                filename = _[0]
            _ = re.findall('.+\.([a-zA-Z]+)', filename)
            if not _:
                break
            fmt = _[0]
            # http://exhentai.org/fullimg.php?gid=577354&page=2&key=af594b7cf3
            _ = re.findall('.+/(\d+)-(\d*)', r._real_url)
            if not _:
                break
            index = _[0]
            fullurl = re.findall('class="mr".+<a href="(.+)"\s*>Download original', r.text)
            fullsize = re.findall('Download\soriginal\s[0-9]+\sx\s[0-9]+\s(.*)\ssource', r.text)  # like 2.20MB
            if fullurl:
                fullurl = util.htmlescape(fullurl[0])
            else:
                fullurl = picurl
            _= re.findall("return nl\('([\d\-]+)'\)", r.text)
            if not _:
                break
            js_nl = _[0]
            reload_url = "%s%snl=%s" % (r._real_url, "&" if "?" in r._real_url else "?", js_nl)
            if ori:
                # we will parse the 302 url to get original filename
                return suc((fullurl, reload_url, filename))
            else:
                return suc((picurl, reload_url, filename))

        return fail((ERR_SCAN_REGEX_FAILED, r._real_url))

    return flt_imgurl

def download_file_wrapper(dirpath):
    @flt_quota_check
    def download_file(r, suc, fail, dirpath = dirpath):
        # input image/archive response
        # return (binary, url) if suc; return (errocode, url) if fail
        if r.status_code == 404:
            return fail((ERR_HATH_NOT_FOUND, r._real_url, r.url))
        p = RE_IMGHASH.findall(r.url)
        # if multiple hash-size-h-w-type is found, use the last one
        # the first is original and the last is scaled
        # _FakeReponse will be filtered in flt_quota_check
        if not r.content_length or \
            p and p[-1] and int(p[-1][1]) != r.content_length:
            return fail((ERR_IMAGE_BROKEN, r._real_url, r.url))
        if not hasattr(r, 'iter_content_cb'):
            return fail((ERR_STREAM_NOT_IMPLEMENTED, r._real_url, r.url))

        # merge the iter_content iterator with our custom stream_cb
        def _yield(chunk_size=16384, _r=r):
            from requests.exceptions import ConnectionError
            length_read = 0
            try:
                for _ in _r.iter_content(chunk_size):
                    length_read += len(_)
                    _r.iter_content_cb(_)
                    yield _
            except ConnectionError: # read timeout
                fail((ERR_IMAGE_BROKEN, r._real_url, r.url))
                raise DownloadAbortedException()
            if length_read != r.content_length:
                fail((ERR_IMAGE_BROKEN, r._real_url, r.url))
                raise DownloadAbortedException()
            
        suc((_yield, r._real_url, r.url))

    return download_file


def reset_quota(r, suc, fail):
    # reset quota response
    # reset quota if suc; finish task if fail
    pass



#!/usr/bin/env python
# coding:utf-8
# constants module
# Contributor:
#      fffonion        <fffonion@gmail.com>

import os
import re
import sys
import locale

PY3K = sys.version_info[0] == 3
IRONPYTHON = sys.platform == 'cli'
EXEBUNDLE = getattr(sys, 'frozen', False)
LOCALE = locale.getdefaultlocale()[0]
CODEPAGE = locale.getdefaultlocale()[1] or 'ascii'
ANDROID = 'ANDROID_ARGUMENT' in os.environ

__version__ = 2.020
DEVELOPMENT = False

SCRIPT_NAME = "xeHentai"

# https://github.com/soimort/you-get/you-get
if getattr(sys, 'frozen', False):
    # The application is frozen
    FILEPATH = os.path.dirname(os.path.realpath(sys.executable))
else:
    # The application is not frozen
    # Change this bit to match where you store your data files:
    FILEPATH = sys.path[0]

DUMMY_FILENAME = "-dummy-"
RENAME_TMPDIR = "-xeh-conflict-"

RE_INDEX = re.compile('.+/(\d+)/([^\/]+)/*')
RE_GALLERY = re.compile('/([a-f0-9]{10})/[^\-]+\-(\d+)')
RE_IMGHASH = re.compile('/([a-f0-9]{40})-(\d+)-(\d+)-(\d+)-([a-z]{,4})')
RE_FULLIMG = re.compile('fullimg.php\?gid=([a-z0-9]+)&page=(\d+)&key=')

__restr_webpage = '^https*://([^\.]+\.)*(?:[g\.]*e-|ex)hentai.org'
RE_URL_WEBPAGE = re.compile(__restr_webpage)
RE_URL_IMAGE = re.compile('(?!%s)' % __restr_webpage)
# matches all
RE_URL_ALL = re.compile('.')

RE_LOCAL_ADDR = re.compile('(^localhost)|(^127\.)|(^192\.168\.)|(^10\.)|(^172\.1[6-9]\.)|(^172\.2[0-9]\.)|(^172\.3[0-1]\.)|(^::1$)|(^[fF][cCdD])')

RESTR_SITE = "https*://(?:[g\.]*e\-|ex)hentai\.org"

FALLBACK_CF_IP = ("104.24.255.11", "104.24.254.11")
FALLBACK_IP_MAP = {
    'e-hentai.org': FALLBACK_CF_IP,
    'forums.e-hentai.org': ("94.100.18.243", ),
    'exhentai.org': ("217.23.13.91","217.23.13.45","109.236.84.136","109.236.92.143","109.236.84.145","109.236.92.166")
} 

DEFAULT_MAX_REDIRECTS = 30

XEH_STATE_RUNNING = 0
XEH_STATE_SOFT_EXIT = 1 # wait until current task finish and exit
XEH_STATE_FULL_EXIT = 2 # finish current task stage and exit
XEH_STATE_CLEAN = 3

TASK_STATE_PAUSED = 0
TASK_STATE_WAITING = 1
TASK_STATE_GET_META = 2
# TASK_STATE_GET_HATHDL = 3
TASK_STATE_SCAN_PAGE = 3
TASK_STATE_SCAN_IMG = 4
TASK_STATE_SCAN_ARCHIVE = 5
TASK_STATE_DOWNLOAD = 10
TASK_STATE_MAKE_ARCHIVE = 19
TASK_STATE_FINISHED = 20
TASK_STATE_FAILED = -1

ERR_NO_ERROR = 0
ERR_URL_NOT_RECOGNIZED = 1000
ERR_CANT_DOWNLOAD_EXH = 1001
ERR_ONLY_VISIBLE_EXH = 1002
ERR_MALFORMED_HATHDL = 1003
ERR_GALLERY_REMOVED = 1004
ERR_IMAGE_RESAMPLED = 1005
ERR_QUOTA_EXCEEDED = 1006
ERR_KEY_EXPIRED = 1007
ERR_NO_PAGEURL_FOUND = 1008
ERR_CONNECTION_ERROR = 1009
ERR_IP_BANNED = 1010
ERR_HATH_NOT_FOUND = 1011
ERR_IMAGE_BROKEN = 1012
ERR_SCAN_REGEX_FAILED = 1013
ERR_STREAM_NOT_IMPLEMENTED = 1013
ERR_TASK_NOT_FOUND = 1101
ERR_SAVE_SESSION_FAILED = 1103
ERR_TASK_LEVEL_UNDEF = 1104
ERR_DELETE_RUNNING_TASK = 1105
ERR_TASK_CANNOT_PAUSE = 1106
ERR_TASK_CANNOT_RESUME = 1107
# ERR_HATHDL_NOTFOUND = 1108
ERR_RPC_UNAUTHORIZED = 1200
ERR_CANNOT_CREATE_DIR = 1300
ERR_CANNOT_MAKE_ARCHIVE = 1301
ERR_NOT_RANGE_FORMAT = 1302
ERR_RPC_PARSE_ERROR = -32700
ERR_RPC_INVALID_REQUEST = -32600
ERR_RPC_METHOD_NOT_FOUND = -32601
ERR_RPC_INVALID_PARAMS = -32602
ERR_RPC_EXEC_ERROR = -32603


class DownloadAbortedException(Exception):
    pass



#!/usr/bin/env python
# coding:utf-8
# Contributor:
#      fffonion        <fffonion@gmail.com>

import re
import math
import time
import random
import requests
from requests.adapters import HTTPAdapter
import traceback
from threading import Thread, RLock
from . import util
from .const import *
from .i18n import i18n
from .proxy import PoolException
if PY3K:
    from queue import Queue, Empty
    from urllib.parse import urlparse, urlunparse
else:
    from Queue import Queue, Empty
    from urlparse import urlparse, urlunparse

# pinfo = {'http':'socks5://127.0.0.1:16963', 'https':'socks5://127.0.0.1:16963'}

class _FakeResponse(object):
    def __init__(self, url):
        self.status_code = 600
        self.content = None
        self.url = self._real_url = url
        self.headers = {}

class FallbackIpAdapter(HTTPAdapter):
    def __init__(self, ip_map=FALLBACK_IP_MAP, **kwargs):
        self.ip_map = ip_map
        kwargs.update({'max_retries': 1})
        requests.adapters.HTTPAdapter.__init__(self, **kwargs)

    # override
    def get_connection(self, url, proxies=None):
        if not proxies:
            parsed = urlparse(url)
            _hostname = parsed.hostname
            _scheme = parsed.scheme
            if _hostname in self.ip_map:
                _parsed = list(parsed)
                # alter the hostname
                _hostname = '%s%s' % (random.choice(self.ip_map[_hostname]),
                                        (":%d" % parsed.port) if parsed.port else "")
                _scheme = 'https'
            return self.poolmanager.connection_from_host(_hostname, parsed.port, scheme=_scheme,
                                                            pool_kwargs={'assert_hostname': parsed.hostname})
        else:
            # fallback
            return requests.adapters.HTTPAdapter.get_connection(self, url, proxies)
    
    def add_headers(self, request, **kwargs):
        if not request.headers.get('Host'):
            parsed = urlparse(request.url)
            request.headers['Host'] = parsed.hostname
    
    def cert_verify(self, conn, url, verify, cert):
        # let the super run verify process
        if url.startswith('http://'):
            url = "https://%s" % url[7:]
        return requests.adapters.HTTPAdapter.cert_verify(self, conn, url, verify, cert)

class HttpReq(object):
    def __init__(self, headers = {}, proxy = None, proxy_policy = None, retry = 10, timeout = 20, logger = None, tname = "main"):
        self.session = requests.Session()
        self.session.headers = headers
        for u in ('forums.e-hentai.org', 'e-hentai.org', 'exhentai.org'):
            self.session.mount('http://%s' % u, FallbackIpAdapter())
            self.session.mount('https://%s' % u, FallbackIpAdapter())
        self.session.mount('http://', HTTPAdapter(max_retries=0))
        self.retry = retry
        self.timeout = timeout
        self.proxy = proxy
        self.proxy_policy = proxy_policy
        self.logger = logger
        self.tname = tname

    def request(self, method, url, _filter, suc, fail, data=None, stream_cb=None):
        retry = 0
        url_history = [url]
        while retry < self.retry:
            try:
                headers = {}
                # if proxy_policy is set and match current url, use proxy
                if self.proxy and self.proxy_policy and self.proxy_policy.match(url):
                    f, __not_good = self.proxy.proxied_request(self.session)
                else:
                    f = self.session.request
                r = f(method, url,
                    allow_redirects=False,
                    data=data,
                    timeout=self.timeout,
                    stream=stream_cb != None)
            except requests.RequestException as ex:
                self.logger.warning("%s-%s %s %s: %s" % (i18n.THREAD, self.tname, method, url, ex))
                time.sleep(random.random() + 0.618)
            else:
                if r.headers.get('content-length'):
                    r.content_length = int(r.headers.get('content-length'))
                elif not stream_cb:
                    r.content_length = len(r.content)
                else:
                    r.content_length = 0
                self.logger.verbose("%s-%s %s %s %d %d" % (i18n.THREAD, self.tname, method, url, r.status_code, r.content_length))
                # if it's a redirect, 3xx
                if r.status_code > 300 and r.status_code < 400:
                    _new_url = r.headers.get("location")
                    if _new_url:
                        url_history.append(url)
                        if len(url_history) > DEFAULT_MAX_REDIRECTS:
                            self.logger.warning("%s-%s %s %s: too many redirects" % (i18n.THREAD, self.tname, method, url))
                            return _filter(_FakeResponse(url_history[0]), suc, fail)
                        url = _new_url
                        continue
                # intercept some error to see if we can change IP
                if self.proxy and r.content_length < 1024 and \
                    re.match("Your IP address has been temporarily banned", r.text):
                    _t = util.parse_human_time(r.text)
                    self.logger.warn(i18n.PROXY_DISABLE_BANNED % _t)
                    # fail this proxy immediately and set expire time
                    __not_good(expire = _t)
                    continue
                
                r.encoding = "utf-8"
                # r._text_bytes = r.text.encode("utf-8")
                r._real_url = url_history[-1]

                r.iter_content_cb = stream_cb

                return _filter(r, suc, fail)
            retry += 1
        return _filter(_FakeResponse(url_history[0]), suc, fail)



class HttpWorker(Thread, HttpReq):
    def __init__(self, tname, task_queue, flt, suc, fail, headers={}, proxy=None, proxy_policy=None,
            retry=3, timeout=10, logger=None, keep_alive=None, stream_mode=False):
        """
        Construct a new 'HttpWorker' obkect

        :param tname: The name of this http worker
        :param task_queue: The task Queue instance
        :param flt: the filter function
        :param suc: the function to call when succeeded
        :param fail: the function to call when failed
        :param headers: custom HTTP headers
        :param proxy: proxy dict
        :param proxy_policy: a function to determine whether proxy should be used
        :param retry: retry count
        :param timeout: timeout in seconds
        :param logger: the Logger instance
        :param keep_alive: the callback to send keep alive
        :param stream_mode: set the request to use stream mode, keep_alive will be called every iteration
        :return: returns nothing
        """
        HttpReq.__init__(self, headers, proxy, proxy_policy, retry, timeout, logger, tname = tname)
        Thread.__init__(self, name = tname)
        Thread.setDaemon(self, True)
        self.task_queue = task_queue
        self.logger = logger
        self._keepalive = keep_alive
        self._exit = lambda x: False
        self.flt = flt
        self.f_suc = suc
        self.f_fail = fail
        self.stream_mode = stream_mode
        # if we don't checkin in this zombie_threshold time, monitor will regard us as zombie
        self.zombie_threshold = timeout * (retry + 1) 
        self.run_once = False

    def _finish_queue(self, *args):
        # exit if current queue is finished
        return self.run_once and self.task_queue.empty()

    def run(self):
        self.logger.verbose("t-%s start" % self.name)
        _stream_cb = None
        if self.stream_mode:
            _stream_cb = lambda x:self._keepalive(self)
        while not self._keepalive(self) and not self._exit(self):
            try:
                url = self.task_queue.get(False)
            except Empty:
                time.sleep(1)
                continue
            self.run_once = True
            try:
                self.request("GET", url, self.flt, self.f_suc, self.f_fail, stream_cb=_stream_cb)
            except PoolException as ex:
                self.logger.warning("%s-%s %s" % (i18n.THREAD, self.tname, str(ex)))
                break
            except Exception as ex:
                self.logger.warning(i18n.THREAD_UNCAUGHT_EXCEPTION % (self.tname, traceback.format_exc()))
                self.flt(_FakeResponse(url), self.f_suc, self.f_fail)
        # notify monitor the last time
        self.logger.verbose("t-%s exit" % self.name)
        self._keepalive(self, _exit = True)

class ArchiveWorker(Thread):
    # this worker is not managed by monitor
    def __init__(self, logger, task, exit_check = None):
        Thread.__init__(self, name = "archiver%s" % task.guid)
        Thread.setDaemon(self, True)
        self.logger = logger
        self.task = task
        self._exit = lambda x: False

    def run(self):
        while self.task.state < TASK_STATE_FINISHED:
            if self._exit(self) or self.task.state in (TASK_STATE_PAUSED, TASK_STATE_FAILED):
                return
            time.sleep(1)
        self.logger.info(i18n.TASK_START_MAKE_ARCHIVE % self.task.guid)
        self.task.state = TASK_STATE_MAKE_ARCHIVE
        t = time.time()
        try:
            pth = self.task.make_archive()
        except Exception as ex:
            self.task.state = TASK_STATE_FAILED
            self.logger.error(i18n.TASK_ERROR % (self.task.guid, i18n.c(ERR_CANNOT_MAKE_ARCHIVE) % traceback.format_exc()))
        else:
            self.task.state = TASK_STATE_FINISHED
            self.logger.info(i18n.TASK_MAKE_ARCHIVE_FINISHED % (self.task.guid, pth, time.time() - t))


class Monitor(Thread):
    def __init__(self, req, proxy, logger, task, exit_check=None, ignored_errors=[]):
        Thread.__init__(self, name = "monitor%s" % task.guid)
        Thread.setDaemon(self, True)
        # the count of votes per error code
        self.vote_result = {}
        # the error code to be ignored
        self.vote_cleared = set().union(ignored_errors)
        self.thread_last_seen = {}
        self.dctlock = RLock()
        self.votelock = RLock()
        self.thread_ref = {}
        self.thread_zombie = set()
        # HttpReq instance
        self.req = req
        # proxy.Pool instance
        self.proxy = proxy
        self.logger = logger
        self.task = task
        self._exit = exit_check if exit_check else lambda x: False
        self._cleaning_up = False
        if os.name == "nt":
            self.set_title = lambda s:os.system("TITLE %s" % (
                s if PY3K else s.encode(CODEPAGE, 'replace')))
        elif os.name == 'posix':
            import sys
            self.set_title = lambda s:sys.stdout.write("\033]2;%s\007" % (
                s if PY3K else s.encode(CODEPAGE, 'replace')))

    def set_vote_ns(self, tnames):
        t = time.time()
        self.thread_last_seen = {k:t for k in tnames}

    def vote(self, tname, code):
        # thread_id, result_code
        self.votelock.acquire()
        if code != ERR_NO_ERROR:
            self.logger.verbose("t-%s vote:%s" % (tname, code))
        if code not in self.vote_result:
            self.vote_result[code] = 1
        else:
            self.vote_result[code] += 1
        self.votelock.release()

    def wrk_keepalive(self, wrk_thread, _exit = False):
        tname = wrk_thread.name
        if tname in self.thread_zombie:
            self.thread_zombie.remove(tname)
        # all image downloaded
        # task is finished or failed
        # monitor is exiting or worker notify its exit
        _ = self.task.meta['finished'] == self.task.meta['total'] or \
            self.task.state in (TASK_STATE_FINISHED, TASK_STATE_FAILED) or \
            self._exit("mon") or _exit
        # self.logger.verbose("mon#%s %s ask, %s, %s" % (self.task.guid, tname, _,
        #    self.thread_last_seen))
        if _ or not wrk_thread.is_alive():
            self.dctlock.acquire()
            if tname in self.thread_last_seen:
                del self.thread_last_seen[tname]
            if tname in self.thread_ref:
                del self.thread_ref[tname]
            self.dctlock.release()
        else:
            self.thread_last_seen[tname] = time.time()
            if tname not in self.thread_ref:
                self.thread_ref[tname] = wrk_thread
        return _

    # def _rescan_pages(self):
    #     # not using
    #     # throw away existing page urls
    #     while True:
    #         try:
    #             self.task.page_q.get(False)
    #         except Empty:
    #             break
    #     # put page into task.list_q
    #     [self.task.list_q.put("%s/?p=%d" % (self.task.url, x)
    #         for x in range(1, 1 + int(math.ceil(self.task.meta['total']/20.0))))
    #     ]
    #     print(self.task.list_q.qsize())

    def _check_vote(self):
        if False and ERR_IMAGE_RESAMPLED in self.vote_result and ERR_IMAGE_RESAMPLED not in self.vote_cleared:
            self.logger.warning(i18n.TASK_START_PAGE_RESCAN % self.task.guid)
            self._rescan_pages()
            self.task.meta['has_ori'] = True
            self.vote_cleared.add(ERR_IMAGE_RESAMPLED)
        elif ERR_QUOTA_EXCEEDED in self.vote_result and \
            ERR_QUOTA_EXCEEDED not in self.vote_cleared and \
            self.vote_result[ERR_QUOTA_EXCEEDED] >= len(self.thread_last_seen):
            self.logger.error(i18n.TASK_STOP_QUOTA_EXCEEDED % self.task.guid)
            self.task.state = TASK_STATE_FAILED

    def run(self):
        CHECK_INTERVAL = 10
        STUCK_INTERVAL = 90
        intv = 0
        self.set_title(i18n.TASK_START % self.task.guid)
        last_change = time.time()
        last_finished = -1
        while len(self.thread_last_seen) > 0:
            intv += 1
            self._check_vote()
            for k in list(self.thread_last_seen.keys()):
                _zombie_threshold = self.thread_ref[k].zombie_threshold if k in self.thread_ref else 30
                if time.time() - self.thread_last_seen[k] > _zombie_threshold:
                    if k not in self.thread_ref and self.thread_ref[k].is_alive():
                        self.logger.warning(i18n.THREAD_MAY_BECOME_ZOMBIE % k)
                        self.thread_zombie.add(k)
                    else:
                        self.logger.warning(i18n.THREAD_SWEEP_OUT % k)
                    del self.thread_last_seen[k]
            if intv == CHECK_INTERVAL:
                _ = "%s %dR/%dZ, %s %dR/%dD" % (
                    i18n.THREAD,
                    len(self.thread_last_seen), len(self.thread_zombie),
                    i18n.QUEUE,
                    self.task.img_q.qsize(),
                    self.task.meta['finished'])
                self.logger.info(_)
                self.set_title(_)
                intv = 0
                # if not downloading any new images in 1.5 min, exit
                if last_finished != self.task.meta['finished']:
                    last_change = time.time()
                    last_finished = self.task.meta['finished']
                else:
                    if time.time() - last_change > STUCK_INTERVAL:
                        self.logger.warning(i18n.TASK_STUCK % self.task.guid)
                        break
            time.sleep(0.5)
        if self.task.meta['finished'] == self.task.meta['total']:
            _err = self.task.rename_fname()
            if _err:
                self.logger.warning(i18n.XEH_RENAME_HAS_ERRORS % (
                    "\n".join(map(lambda x:"%s => %s : %s" % x, _err))
                ))
            self.set_title(i18n.TASK_FINISHED % self.task.guid)
            self.logger.info(i18n.TASK_FINISHED % self.task.guid)
            self.task.state = TASK_STATE_FINISHED
        self.task.cleanup()

if __name__ == '__main__':
    print(HttpReq().request("GET", "https://ipip.tk", lambda x:x, None, None))



# coding:utf-8
# DO NOT EDIT THIS FILE
# make a copy to your working directory
# and edit that file

# Daemon mode
daemon = False

# set download directory
dir = "."
# download original images, needs to login
download_ori = False
# Set if use Japanese title if available
jpn_title = True
# rename gallery image to original name, use sequence name if turned off
rename_ori = False

# set download proxies
# currenlty supported: socks5/4a, http(s), glype
# by default, proxy is only used on webpages
proxy = []
# also use proxy to download images
proxy_image = True
# only use proxy on images, not webpages
# if set to True, the value of proxy_image will be ignored
proxy_image_only = False

# bind jsonrpc server to this address
rpc_interface = 'localhost'
# bind jsonrpc server to this port
rpc_port = None
# jsonrpc secret string
rpc_secret = None

# make an archive (.zip) after download and delete directory
make_archive = False
# specify ranges of images to be downloaded, in format
# start-end, or single index, use comma to concat
# multiple ranges, e.g.: 5-10,15,20-25, default to
# download all images
download_range = None
# scan threads count
scan_thread_cnt = 1
# download threads count
download_thread_cnt = 5
# set image download timeout
download_timeout = 10

# ignore these error codes, continue download
# to use predefined error codes, use:
# import const as __c
# ignored_errors = [__c.ERR_QUOTA_EXCEEDED]
ignored_errors = []

# define log path
log_path = "eh.log"
# set log level
log_verbose = 2

# save tasks to h.json
save_tasks = False

# delete files when deleting a task
delete_task_files = False




#!/usr/bin/env python
# coding:utf-8
# Contributor:
#      fffonion        <fffonion@gmail.com>

import re
import time
import json
import zipfile
import traceback
from hashlib import md5
from threading import Thread
from .const import *
from .const import __version__
from .i18n import i18n
if PY3K:
    from socketserver import ThreadingMixIn
    from http.server import HTTPServer, BaseHTTPRequestHandler
    from io import IOBase
    from io import BytesIO as StringIO
    from urllib.parse import urlparse
else:
    from SocketServer import ThreadingMixIn
    from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler
    from cStringIO import StringIO 
    from urlparse import urlparse

cmdre = re.compile("([a-z])([A-Z])")
pathre = re.compile("/(?:jsonrpc|img|zip)")
imgpathre = re.compile("/img")
zippathre = re.compile("/zip")

class RPCServer(Thread):
    def __init__(self, xeH, bind_addr, secret = None, logger = None, exit_check = None):
        Thread.__init__(self, name = "rpc")
        Thread.setDaemon(self, True)
        self.xeH = xeH
        self.bind_addr = bind_addr
        self.secret = secret
        self.logger = logger
        self.server = None
        self._exit = exit_check if exit_check else lambda x:False

    def run(self):
        try:
            self.server = ThreadedHTTPServer(self.bind_addr, lambda *x: Handler(self.xeH, self.secret, *x))
        except Exception as ex:
            self.logger.error(i18n.RPC_CANNOT_BIND % traceback.format_exc())
        else:
            self.logger.info(i18n.RPC_STARTED % (self.bind_addr[0], self.bind_addr[1]))
            while not self._exit("rpc"):
                self.server.handle_request()

def is_file_obj(obj):
    if PY3K:
        return isinstance(obj, IOBase)
    return isinstance(obj, file)

def is_str_obj(obj):
    if PY3K:
        return isinstance(obj, str)
    return isinstance(obj, basestring)

def hash_link(secret, url):
    _ = "%s-xehentai-%s" % (secret if secret else "", url)
    if PY3K:
        _ = _.encode('utf-8')
    return md5(_).hexdigest()[:8]

def gen_thumbnail(fh, args):
    # returns a new file handler if resized
    # and a boolean indicates there'e error
    try:
        from PIL import Image
    except:
        return fh, True
    if 'w' not in args and 'h' not in args:
        return fh, False
    size = (int(args['w']) if 'w' in args else int(args['h']),
            int(args['h']) if 'h' in args else int(args['w']))
    if not is_file_obj(fh):
        fh = StringIO(fh)
    with Image.open(fh) as img:
        img.thumbnail(size)
        ret_fh = StringIO()
        img.save(ret_fh, format=img.format)
        ret = ret_fh.getvalue()
        ret_fh.close()
        fh.close()
        return ret, False
    
def jsonrpc_resp(request, ret = None, error_code = None, error_msg = None):
    r = {
        "id":None if not request["id"] else request["id"],
        "jsonrpc":"2.0",
    }
    if error_code:
        r['error'] = {
            'code':error_code,
            "message":i18n.c(error_code) if not error_msg else error_msg
        }
    else:
        r['result'] = ret
    return json.dumps(r)

def path_filter(func):
    def f(self):
        if not pathre.match(self.path):
            self.send_response(404)
            self.send_header("Access-Control-Allow-Origin", "*")
            self.end_headers()
            self.wfile.write(b'\n')
            return
        func(self)
    return f

class Handler(BaseHTTPRequestHandler):

    def __init__(self, xeH, secret, *args):
        self.secret = secret
        self.args = args
        self.xeH = xeHentaiRPCExtended(xeH, secret)
        BaseHTTPRequestHandler.__init__(self, *args)

    def version_string(self):
        return "xeHentai/%s" % __version__
    
    def serve_file(self, f):
        f.seek(0, os.SEEK_END)
        size = f.tell()
        self.xeH.logger.verbose("GET %s 200 %d %s" % (self.path, size, self.client_address[0]))
        self.send_header("Content-Length", size)
        f.seek(0, os.SEEK_SET)
        self.end_headers()
        while True:
            buf = f.read(51200)
            if not buf:
                break
            self.wfile.write(buf)
        return size

    def do_OPTIONS(self):
        self.send_response(200)
        self.send_header("Access-Control-Allow-Origin", "*")
        self.send_header("Access-Control-Allow-Headers", "Content-Type")
        self.send_header("Access-Control-Allow-Methods", "POST, GET, OPTIONS")
        self.send_header("Access-Control-Max-Age", "1728000")
        self.end_headers()
        self.wfile.write(b'\n')

    @path_filter
    def do_GET(self):
        code = 200
        rt = b''
        mime = "text/html"
        while True:
            if imgpathre.match(self.path):
                args = dict(q.split("=") for q in urlparse(self.path).query.split("&") if q)
                _ = urlparse(self.path).path.split("/")
                if len(_) < 5:
                    code = 400
                    break
                _, _, _hash, guid, fid = _[:5]
                right_hash = hash_link(self.secret, "%s/%s" % (guid, fid))
                if right_hash != _hash:
                    self.xeH.logger.warning("RPC: hash mismatch %s != %s" % (right_hash, _hash))
                    code = 403
                    break
                path, f, mime = self.xeH._get_image_path(guid, fid)
                if not f or not os.path.exists(os.path.join(path, f)):
                    zipf = "%s.zip" % path
                    if not os.path.exists(zipf):
                        self.xeH.logger.warning("RPC: can't find %s" % f)
                        code = 404
                        break
                    else:
                        z = zipfile.ZipFile(zipf)
                        try:
                            rt = z.read(f)
                        except Exception as ex:
                            self.xeH.logger.warning("RPC: can't find %s in zipfile: %s" % (f, ex))
                            code = 404
                            break
                        z.close()
                else:
                    rt = open(os.path.join(path, f), 'rb')
                rt, _error = gen_thumbnail(rt, args)
                if _error:
                    self.xeH.logger.warning("RPC: PIL needed for generating thumbnail")
            elif zippathre.match(self.path):
                # args = urlparse(_).query
                _ = urlparse(self.path).path.split("/")
                if len(_) < 5:
                    code = 400
                    break
                _, _, _hash, guid, fname = _[:5]
                fname = fname.split('?')[0]
                right_hash = hash_link(self.secret, "%s" % guid)
                if right_hash != _hash:
                    self.xeH.logger.warning("RPC: hash mismatch %s != %s" % (right_hash, _hash))
                    code = 403
                    break
                f = self.xeH._get_archive_path(guid)
                mime = 'application/zip'
                if not f or not os.path.exists(f):
                    self.xeH.logger.warning("RPC: can't find %s" % f)
                    code = 404
                    break
                rt = open(f, 'rb')
            else:
                # fallback to rpc request
                rt = jsonrpc_resp({"id":None}, error_code = ERR_RPC_INVALID_REQUEST)
                mime = "application/json-rpc"
            break

        self.send_response(code)
        self.send_header("Access-Control-Allow-Origin", "*")
        self.send_header("Content-Type", mime)
        
        if is_file_obj(rt):
            size = self.serve_file(rt)
            rt.close()
        else:
            self.xeH.logger.verbose("GET %s 200 %d %s" % (self.path, len(rt), self.client_address[0]))
            self.send_header("Content-Length", len(rt))
            self.end_headers()
            self.wfile.write(rt)
        self.wfile.write(b'\n')
        return

    @path_filter
    def do_POST(self):
        _get_header = lambda h: self.headers.get_all(h)[0] if PY3K else \
            self.headers.getheader(h)
        d = self.rfile.read(int(_get_header('Content-Length')))
        code = 200
        rt = b''
        while True:
            try:
                if PY3K:
                    d = d.decode('utf-8')
                j = json.loads(d)
                assert('method' in j and j['method'] != None and 'id' in j)
            except ValueError:
                code = 400
                rte = jsonrpc_resp({"id":None}, error_code = ERR_RPC_PARSE_ERROR)
                break
            except AssertionError:
                code = 400
                rt = jsonrpc_resp({"id":None}, error_code = ERR_RPC_INVALID_REQUEST)
                break
            cmd = re.findall("xeH\.(.+)", j['method'])
            if not cmd:
                code = 404
                rt = jsonrpc_resp({"id":j['id']}, error_code = ERR_RPC_METHOD_NOT_FOUND)
                break
            # let's make fooBar to foo_bar
            cmd_r = cmdre.sub(lambda m: "%s_%s" % (m.group(1), m.group(2).lower()), cmd[0])
            if not hasattr(self.xeH, cmd_r) or cmd_r.startswith("_"):
                code = 404
                rt = jsonrpc_resp({"id":j['id']}, error_code = ERR_RPC_METHOD_NOT_FOUND)
                break
            params = ([], {}) if 'params' not in j else j['params']
            if self.secret:
                authorized = False
                while True:
                    if len(params[0]) == 0:
                        break
                    secret = params[0][0]
                    if not PY3K and isinstance(secret, unicode):
                        secret = secret.encode('utf-8')
                    if is_str_obj(secret) and re.findall("token:%s" % self.secret, secret):
                        params[0].pop(0)
                        authorized = True
                    break
                if not authorized:
                    code = 403
                    rt = jsonrpc_resp({"id":j['id']}, error_code = ERR_RPC_UNAUTHORIZED)
                    break
            self.xeH.logger.verbose("RPC from: %s, cmd: %s, params: %s" % (self.client_address[0], cmd, params))
            try:
                cmd_rt = getattr(self.xeH, cmd_r)(*params[0], **params[1])
            except (ValueError, TypeError) as ex:
                self.xeH.logger.verbose("RPC exec error:\n%s" % traceback.format_exc())
                code = 500
                rt = jsonrpc_resp({"id":j['id']}, error_code = ERR_RPC_EXEC_ERROR,
                error_msg = str(ex))
                break
            if cmd_rt[0] > 0:
                rt = jsonrpc_resp({"id":j['id']}, error_code = cmd_rt[0], error_msg = cmd_rt[1])
            else:
                rt = jsonrpc_resp({"id":j['id']}, ret = cmd_rt[1])
            break
        self.send_response(code)
        self.send_header("Access-Control-Allow-Origin", "*")
        self.send_header("Content-Type", "application/json-rpc")
        self.send_header("Content-Length", len(rt))
        self.end_headers()
        if PY3K:
            rt = rt.encode('utf-8')
        self.wfile.write(rt)
        self.wfile.write(b'\n')
        return


    def log_message(self, format, *args):
        return

# extend xeHentai class for rpc commands
class xeHentaiRPCExtended(object):
    def __init__(self, xeH, secret):
        self.xeH = xeH
        self.secret = secret
    
    def get_info(self):
        ret = {"version": self.verstr,
            "threads_zombie": 0, "threads_running": 0,
            "queue_pending": 0, "queue_finished": 0
        }
        if hasattr(self, '_monitor'):
            ret['threads_running'] = len(self._monitor.thread_last_seen)
            ret['threads_zombie'] = len(self._monitor.thread_zombie)
            if self._monitor.task.state > TASK_STATE_PAUSED and self._monitor.task.img_q:
                ret['queue_pending'] = self._monitor.task.img_q.qsize()
                ret['queue_finished'] = self._monitor.task.meta['finished']
            else:
                ret['queue_pending'] = 0
                ret['queue_finished'] = 0
        return ERR_NO_ERROR, ret
    
    def get_config(self):
        rt = {k: v for k, v in self.cfg.items() if not k.startswith('rpc_') and k not in ('urls',)}
        return ERR_NO_ERROR, rt
    
    def update_config(self, **cfg_dict):
        cfg_dict = {k: v for k, v in cfg_dict.items() if not k.startswith('rpc_') and k not in ('urls',)}
        if 'proxy' in cfg_dict:
            self.xeH.update_config(**cfg_dict)
        return self.get_config()
           
    def list_tasks(self, level = "download"):
        reverse_mode = False
        if level.startswith('!'):
            reverse_mode = True
            level = level[1:]
        level = "TASK_STATE_%s" % level.upper()
        if level not in globals():
            return ERR_TASK_LEVEL_UNDEF, None
        lv = globals()[level]
        rt = [{_k:_v for _k, _v in v.to_dict().items() if _k not in
            ('reload_map', 'filehash_map', 'renamed_map', 'img_q', 'page_q')}
                 for _, v in self._all_tasks.items() if 
                    (reverse_mode and v.state != lv) or (not reverse_mode and v.state == lv)]
        return ERR_NO_ERROR, rt
    
    def _get_image_path(self, guid, fid):
        mime_map = {
            "jpg": "image/jpeg",
            "jepg": "image/jpeg",
            "png": "image/png",
            "gif": "image/gif",
            "bmp": "image/bmp",
            "webp": "image/webp"
        }
        if guid not in self._all_tasks:
            return None, None, None
        t = self._all_tasks[guid]
        fid = str(fid)
        if fid in t.renamed_map:
            f = t.renamed_map[fid]
        else:
            f = t.get_fidpad(fid)

        ext = os.path.splitext(f)[1].lower()[1:]
        if ext not in mime_map:
            mime = "application/octet-stream"
        else:
            mime = mime_map[ext]
        return t.get_fpath(), f, mime
    
    def _get_archive_path(self, guid):
        if guid not in self._all_tasks:
            return None, None
        t = self._all_tasks[guid]
        st = time.time()
        pth = t.make_archive(False)
        et = time.time()
        if et - st > 0.1:
            self.logger.warning('RPC: %.2fs taken to get archive' % (et - st))
        return pth
    
    def get_image(self, guid, request_range=None):
        if guid not in self._all_tasks:
            return ERR_TASK_NOT_FOUND, None
        t = self._all_tasks[guid]
        start = 1
        end = t.meta['total'] + 1
        if request_range:
            request_range = str(request_range)
            _ = request_range.split(',')
            if len(_) == 1:
                start = int(request_range)
            else:
                start = int(_[0])
            end = int(_[0]) + 1
        rt = []
        for fid in range(start, end):
            if fid in t.renamed_map:
                f = t.renamed_map[fid]
            else:
                f = t.get_fidpad(fid)
            uri = "%s/%s" % (t.guid, fid)
            rt.append('/img/%s/%s/%s' % (hash_link(self.secret, uri), uri, f))
        return ERR_NO_ERROR, rt

    
    def __getattr__(self, k):
        # fallback attribute handler
        return getattr(self.xeH, k)

class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
    """Handle requests in a separate thread."""
    pass



# coding:utf-8

from ..const import *

err_msg = {
    ERR_URL_NOT_RECOGNIZED: "網址不夠紳士",
    ERR_CANT_DOWNLOAD_EXH: "需要登錄後才能下載里站",
    ERR_ONLY_VISIBLE_EXH: "這個本子只有里站能看到",
    ERR_MALFORMED_HATHDL: "hathdl文件有貓餅，解析失敗",
    ERR_GALLERY_REMOVED: "這個本子被移除了，大概里站能看到",
    ERR_KEY_EXPIRED: "下載鏈接不太正常",
    ERR_NO_PAGEURL_FOUND: "沒有找到頁面鏈接，網站改版了嘛？",
    ERR_CONNECTION_ERROR: "連接有問題？",
    ERR_IP_BANNED: "IP被ban了, 恢復時間: %s",
    ERR_IMAGE_BROKEN: "下載的圖片有貓餅",
    ERR_QUOTA_EXCEEDED: "配額超限",
    ERR_TASK_NOT_FOUND: "沒有該GUID對應的任務",
    ERR_TASK_LEVEL_UNDEF: "任務過濾等級不存在",
    ERR_DELETE_RUNNING_TASK: "無法刪除運行中的任務",
    ERR_TASK_CANNOT_PAUSE: "這個任務無法被暫停",
    ERR_TASK_CANNOT_RESUME: "這個任務無法被恢復",
    ERR_CANNOT_CREATE_DIR: "無法創建文件夾 %s",
    ERR_CANNOT_MAKE_ARCHIVE: "無法製作壓縮包 %s",
    ERR_NOT_RANGE_FORMAT: "'%s'不符合範圍的格式, 正確的格式為 1-3 或者 5",
#    ERR_HATHDL_NOTFOUND: "hathdl文件未找到"
    ERR_RPC_PARSE_ERROR: "Parse error.",
    ERR_RPC_INVALID_REQUEST: "Invalid request.",
    ERR_RPC_METHOD_NOT_FOUND: "Method not found.",
    ERR_RPC_INVALID_PARAMS: "Invalid method parameter(s).",
    ERR_RPC_UNAUTHORIZED: "Unauthorized",
    ERR_RPC_EXEC_ERROR: "",
    ERR_SAVE_SESSION_FAILED: "",
}

ERR_NOMSG = "未指定的錯誤，錯誤號 %d"

XEH_OPT_DESC = "紳♂士下載器"
XEH_OPT_EPILOG = "如果參數未指定，則使用config.py中的默認值; " \
        "討論和反饋問題：https://yooooo.us/2013/xehentai"
XEH_OPT_URLS = "下載頁的網址"
XEH_OPT_u = "用戶名"
XEH_OPT_k = "密碼"
XEH_OPT_c = "Cookie字符串，如果指定了用戶名和密碼，此項會被忽略"
XEH_OPT_o = "是否下載原始圖片（如果存在），需要登錄 (當前: %(default)s)"
XEH_OPT_t = "下載線程數 (當前: %(default)d)"
# XEH_OPT_f = "快速掃描，從hathdl猜測頁面鏈接，但有時會抽風 (當前: %(default)s)"
XEH_OPT_l = "保存日誌的路徑 (當前: %(default)s)"
XEH_OPT_p = "設置代理, 可以指定多次, 當前支持的類型: socks5/4a, http(s), glype. 代理默認只用於掃描網頁 (當前: %(default)s)"
XEH_OPT_proxy_image = "同時使用代理來下載圖片和掃描網頁（當前: %(default)s)"
XEH_OPT_proxy_image_only = "僅使用代理來下載圖片, 不用於掃描網頁 (當前: %(default)s)"
XEH_OPT_d = "設置下載目錄 (當前: %(default)s)"
XEH_OPT_v = "設置日誌裝逼等級 (當前: %(default)s)"
XEH_OPT_i = "交互模式，如果開啟後台模式，此項會被忽略 (當前: %(default)s)"
XEH_OPT_r = "將圖片重命名為原始名稱，如果關閉則使用序號 (當前: %(default)s)"
XEH_OPT_daemon = "後台模式 (當前: %(default)s)"
XEH_OPT_rpc_interface = "設置JSON-RPC監聽IP (當前: %(default)s)"
XEH_OPT_rpc_port = "設置JSON-RPC監聽埠 (當前: %(default)s)"
XEH_OPT_rpc_secret = "設置JSON-RPC密鑰 (當前: %(default)s)"
XEH_OPT_a = "下載完成後生成zip壓縮包並刪除下載目錄 (當前: %(default)s)"
XEH_OPT_delete_task_files = "刪除任務時同時刪除下載的文件 (current: %(default)s)"
XEH_OPT_j = "使用日語標題, 如果關閉則使用英文或羅馬字標題 (當前: %(default)s)"
XEH_OPT_download_range = "設置下載的圖片範圍, 格式為 開始位置-結束位置, 或者單張圖片的位置, " \
"使用逗號來分隔多個範圍, 例如 5-10,15,20-25, 默認為下載所有"
XEH_OPT_timeout = "設置下載圖片的超時 (當前: %(default)s秒)"
XEH_OPT_f = "忽略配額判斷，繼續下載 (當前: %(current)s)"
XEH_OPT_h = "顯示本幫助信息"
XEH_OPT_version = "顯示版本信息"
XEH_OPT_IGNORING_I = "後台模式已忽略 -i 參數"


PS_LOGIN = "當前沒有登陸，要登陸嗎 (y/n)? > "
PS_USERNAME = "輸入用戶名 > "
PS_PASSWD = "輸入密碼   > "
PS_URL = "輸入地址（使用,分割下載多個）> "
PS_PROXY = "輸入代理地址 (可選) > "
PS_DOWNLOAD_ORI = "是否下載原圖（默認%s） (y/n)? > "
PS_RENAME_ORI  = "是否自動重命名（默認%s） (y/n)? > "
PS_MAKE_ARCHIVE = "是否製作zip壓縮包（默認%s） (y/n)? > "
PS_JPN_TITLE = "是否使用日語標題（默認%s） (y/n)? > "
PS_DOWNLOAD_RANGE = "下載範圍, 使用逗號分割多個範圍, 回車下載全部 > "
PS_DOWNLOAD_DIR = "下載目錄 (當前: %s)\n回車確認或輸入新路徑 > "

PROXY_CANDIDATE_CNT = "代理池中有%d個代理"

TASK_PUT_INTO_WAIT = "任務 #%s 已存在, 加入等待隊列"
TASK_ERROR = "任務 #%s 發生錯誤: %s"
TASK_MIGRATE_EXH = "任務 #%s 使用里站地址重新下載"
TASK_TITLE = "任務 #%s 標題 %s"
TASK_WILL_DOWNLOAD_CNT = "任務 #%s 將下載%d個文件，共%d個 "
TASK_START = "任務 #%s 開始"
TASK_FINISHED = "任務 #%s 下載完成"
TASK_START_PAGE_RESCAN = "任務 #%s 圖片被縮放，進行完整掃描"
# TASK_FAST_SCAN = "任務 #%s 使用快速掃描"
TASK_START_MAKE_ARCHIVE = "任務 #%s 開始打包"
TASK_MAKE_ARCHIVE_FINISHED = "任務 #%s 打包完成，保存在: %s, 用時%.1f秒"
TASK_STOP_QUOTA_EXCEEDED = "任務 #%s 配額超限"
TASK_STUCK = "任務 #%s 卡住了, 可能是腳本有bug, 或者網絡連接太慢了"

XEH_STARTED = "xeHentai %s 已啟動"
XEH_LOOP_FINISHED = "程序循環已完成"
XEH_LOGIN_EXHENTAI = "登錄紳士"
XEH_LOGIN_OK = "已成為紳士"
XEH_LOGIN_FAILED = "無法登錄紳士；檢查輸入是否有誤或者換一個帳號。\n推薦在瀏覽器登錄後使用RPC複製cookie到xeHentai (教程: http://t.cn/Rctr4Pf)"
XEH_LOAD_TASKS_CNT = "從存檔中讀取了%d個任務"
XEH_LOAD_OLD_COOKIE = "從1.x版cookie文件從讀取了登錄信息"
XEH_DAEMON_START = "後台進程已啟動，PID為%d"
XEH_PLATFORM_NO_DAEMON = "後台模式不支持您的系統: %s"
XEH_CLEANUP = "擦乾淨..."
XEH_CRITICAL_ERROR = "xeHentai 抽風啦:\n%s"
XEH_DOWNLOAD_ORI_NEED_LOGIN = "下載原圖需要登錄"
XEH_FILE_DOWNLOADED = "紳士-{} 已下載圖片 #{} {}"
XEH_RENAME_HAS_ERRORS = "部分圖片重命名失敗:\n%s"
XEH_DOWNLOAD_HAS_ERROR = "紳士-%s 下載圖片時出錯: %s, 將在稍後重試"

RPC_STARTED = "RPC伺服器監聽在 %s:%d"
RPC_TOO_OPEN = "RPC伺服器監聽在公網IP (%s)，為了安全起見應該設置rpc_secret"
RPC_CANNOT_BIND = "RPC伺服器無法啟動：%s"

SESSION_LOAD_EXCEPTION = "讀取存檔時遇到錯誤: %s"
SESSION_WRITE_EXCEPTION = "寫入存檔時遇到錯誤: %s"

THREAD = "紳士"
THREAD_UNCAUGHT_EXCEPTION = "紳士-%s 未捕獲的異常\n%s"
THREAD_MAY_BECOME_ZOMBIE = "紳士-%s 可能變成了喪屍"
THREAD_SWEEP_OUT = "紳士-%s 掛了, 不再理它"

QUEUE = "隊列"

PROXY_DISABLE_BANNED = "禁用了一個被ban的代理，將在約%s秒後恢復"



# coding:utf-8

from ..const import *

err_msg = {
    ERR_URL_NOT_RECOGNIZED: "url not recognized",
    ERR_CANT_DOWNLOAD_EXH: "can't download exhentai.org without login",
    ERR_ONLY_VISIBLE_EXH: "this gallery is only visible in exhentai.org",
    ERR_MALFORMED_HATHDL: "malformed .hathdl, can't parse",
    ERR_GALLERY_REMOVED: "this gallery has been removed, may be visible in exhentai",
    ERR_KEY_EXPIRED: "image url is expired",
    ERR_NO_PAGEURL_FOUND: "no page url found, change of site structure?",
    ERR_CONNECTION_ERROR: "a connection problem occurs",
    ERR_IP_BANNED: "IP has been banned, retry in %s",
    ERR_IMAGE_BROKEN: "downloaded image is broken",
    ERR_QUOTA_EXCEEDED: "quota exceeded",
    ERR_TASK_NOT_FOUND: "no such task guid",
    ERR_TASK_LEVEL_UNDEF: "task filter level unknown",
    ERR_DELETE_RUNNING_TASK: "can't delete a running task",
    ERR_TASK_CANNOT_PAUSE: "this task can't be paused",
    ERR_TASK_CANNOT_RESUME: "this task can't be resumed",
    ERR_CANNOT_CREATE_DIR: "can't create directory %s",
    ERR_CANNOT_MAKE_ARCHIVE: "can't make archive %s",
    ERR_NOT_RANGE_FORMAT: "'%s' is not a range format, expecting '1-2' or '3'",
#    ERR_HATHDL_NOTFOUND: "hathdl not found",
    ERR_RPC_PARSE_ERROR: "Parse error.",
    ERR_RPC_INVALID_REQUEST: "Invalid request.",
    ERR_RPC_METHOD_NOT_FOUND: "Method not found.",
    ERR_RPC_INVALID_PARAMS: "Invalid method parameter(s).",
    ERR_RPC_UNAUTHORIZED: "Unauthorized",
    ERR_RPC_EXEC_ERROR: "",
    ERR_SAVE_SESSION_FAILED: "",
}

ERR_NOMSG = "undefined error message with code %d"

XEH_OPT_DESC = "xeHentai Downloader NG"
XEH_OPT_EPILOG = "Values shown as current is read from config.py " \
                "and can be overriden by command line options. " \
                "Discuss and bug reporting at https://yooooo.us/2013/xehentai"
XEH_OPT_URLS = "gallery url(s) to download"
XEH_OPT_u = "username"
XEH_OPT_k = "password"
XEH_OPT_c = "cookie string, will be overriden if given -u and -k"
XEH_OPT_o = "download original images, needs to login (current: %(default)s)"
XEH_OPT_t = "download threads count (current: %(default)d)"
# XEH_OPT_f = "fast scan, guess page url from .hathdl file, not working everytime (current: %(default)s)"
XEH_OPT_l = "define log path (current: %(default)s)"
XEH_OPT_p = "set download proxies, can be used multiple times, currenlty supported: socks5/4a, http(s), glype. " \
"Proxies are only used on webpages by default (current: %(default)s)"
XEH_OPT_proxy_image = "use proxies on images and webpages (current: %(default)s)"
XEH_OPT_proxy_image_only = "only use proxies on images, not webpages (current: %(default)s)"
XEH_OPT_d = "set download directory (current: %(default)s)"
XEH_OPT_v = "show more detailed log (current: %(default)s)"
XEH_OPT_i = "interactive mode, will be ignored in daemon mode (current: %(default)s)"
XEH_OPT_r = "rename gallery image to original name, use sequence name if turned off  (current: %(default)s)"
XEH_OPT_daemon = "daemon mode, can't use with -i (current: %(default)s)"
XEH_OPT_rpc_interface = "bind jsonrpc server to this address (current: %(default)s)"
XEH_OPT_rpc_port = "bind jsonrpc server to this port (current: %(default)s)"
XEH_OPT_rpc_secret = "jsonrpc secret string (current: %(default)s)"
XEH_OPT_a = "make an archive (.zip) after download and delete directory (current: %(default)s)"
XEH_OPT_delete_task_files = "delete downloaded files when deleting a task (current: %(default)s)"
XEH_OPT_j = "use Japanese title, use English/Romaji title if turned off (current: %(default)s)"
XEH_OPT_download_range = "specify ranges of images to be downloaded, in format start-end, or single index, " \
"use comma to concat multiple ranges, e.g.: 5-10,15,20-25, default to download all images"
XEH_OPT_timeout = "set image download timeout (current: %(default)ss)"
XEH_OPT_f = "download regardless of quota exceeded warning (current: %(current)s)"
XEH_OPT_h = "show this help message and exit"
XEH_OPT_version = "show program's version number and exit"
XEH_OPT_IGNORING_I = "ignoring -i option in daemon mode"

PS_LOGIN = "login to exhentai (y/n)? > "
PS_USERNAME = "Username > "
PS_PASSWD = "Password > "
PS_URL = "URL (seperate with ,)> "
PS_PROXY = "Proxy (optional) > "
PS_DOWNLOAD_ORI = "Download original (y/n, default:%s)? > "
PS_RENAME_ORI = "Rename to original name (y/n, default:%s)? > "
PS_MAKE_ARCHIVE = "Make archive (y/n, default:%s)? > "
PS_JPN_TITLE = "Use Japanese title (y/n, default:%s)? > "
PS_DOWNLOAD_RANGE = "Download range, press enter to download all > "
PS_DOWNLOAD_DIR = "Download to (default: %s)\npress enter or enter new > "

PROXY_CANDIDATE_CNT = "proxy pool has %d candidates"

TASK_PUT_INTO_WAIT = "task #%s already exists, put into waiting state"
TASK_ERROR = "task #%s error: %s"
TASK_MIGRATE_EXH = "task #%s migrate to exhentai.org"
TASK_TITLE = "task #%s title %s"
TASK_WILL_DOWNLOAD_CNT = "task #%s will download %d/%d files"
TASK_START = "task #%s start"
TASK_FINISHED = "task #%s download finishd"
TASK_START_PAGE_RESCAN = "task #%s resample detected, start full scan"
# TASK_FAST_SCAN = "task #%s uses fast scan"
TASK_START_MAKE_ARCHIVE = "task #%s start making archive"
TASK_MAKE_ARCHIVE_FINISHED = "task #%s archive saved at: %s, use %.1fs"
TASK_STOP_QUOTA_EXCEEDED = "task #%s quota exceeded"
TASK_STUCK = "task #%s is stuck, there may be some bugs in xeHentai, or the connection is too slow"

XEH_STARTED = "xeHentai %s started."
XEH_LOOP_FINISHED = "application task loop finished"
XEH_LOGIN_EXHENTAI = "login exhentai"
XEH_LOGIN_OK = "login exhentai successfully"
XEH_LOGIN_FAILED = "can't login exhentai, check your credentials or try another account.\nIt's recommended to login in browser and use RPC to transfer cookie to xeHentai (see http://t.cn/Rctr4Pf)"
XEH_LOAD_TASKS_CNT = "load %d tasks from saved session"
XEH_LOAD_OLD_COOKIE = "load cookie from legacy cookie file"
XEH_DAEMON_START = "daemon start at PID %d"
XEH_PLATFORM_NO_DAEMON = "daemon mode is not supported on platform: %s"
XEH_CLEANUP = "cleaning up..."
XEH_CRITICAL_ERROR = "xeHentai throws critical error:\n%s"
XEH_DOWNLOAD_ORI_NEED_LOGIN = "haven't login, so I won't download original images"
XEH_FILE_DOWNLOADED = "file downloaded by thread-{} #{} {}"
XEH_RENAME_HAS_ERRORS = "some files are not renamed:\n%s"
XEH_DOWNLOAD_HAS_ERROR = "thread-%s retry because of error: %s"

RPC_STARTED = "RPC server listening on %s:%d"
RPC_TOO_OPEN = "RPC server is listening on public interface (%s) but no rpc_secret defined, which is not safe"
RPC_CANNOT_BIND = "RPC server can't listen on requested address: %s"

SESSION_LOAD_EXCEPTION = "exception occurs when loading saved session: %s"
SESSION_WRITE_EXCEPTION = "exception occurs when writing saved session: %s"

THREAD = "thread"
THREAD_UNCAUGHT_EXCEPTION = "thread-%s uncaught exception\n%s"
THREAD_MAY_BECOME_ZOMBIE = "thread-%s may became zombie"
THREAD_SWEEP_OUT = "thread-%s is dead, deref it"

QUEUE = "queue"

PROXY_DISABLE_BANNED = "disable a banned proxy, expire in about %ss"



#!/usr/bin/env python
# coding:utf-8
# Contributor:
#      fffonion        <fffonion@gmail.com>

from __future__ import absolute_import
import os
import time
import argparse
import traceback
from threading import Thread
from .i18n import i18n
from .core import xeHentai
from .const import *
from .const import __version__
from .util import logger

from . import config as default_config
sys.path.insert(1, FILEPATH)
try:
    import config
except ImportError:
    config = default_config
sys.path.pop(1)

def start():
    opt = parse_opt()
    xeH = xeHentai()
    if opt.daemon:
        if opt.interactive:
            xeH.logger.warning(i18n.XEH_OPT_IGNORING_I)
        if os.name == "posix":
            pid = os.fork()
            if pid == 0:
                sys.stdin.close()
                sys.stdout = open("/dev/null", "w")
                sys.stderr = open("/dev/null", "w")
                return main(xeH, opt)
        elif os.name == "nt":
            return xeH.logger.error(i18n.XEH_PLATFORM_NO_DAEMON % os.name)
        else:
            return xeH.logger.error(i18n.XEH_PLATFORM_NO_DAEMON % os.name)
        xeH.logger.info(i18n.XEH_DAEMON_START % pid)
    else:
        main(xeH, opt)

def main(xeH, opt):
    xeH.update_config(**vars(opt))
    log = xeH.logger
    log.info(i18n.XEH_STARTED % xeH.verstr)
    if opt.cookie:
        xeH.set_cookie(opt.cookie)
    if opt.username and opt.key and not xeH.has_login:
        xeH.login_exhentai(opt.username, opt.key)
    if opt.interactive and not opt.daemon:
        try:
            r = interactive(xeH)
            opt.__dict__.update(r)
            xeH.update_config(**r)
        except (KeyboardInterrupt, SystemExit):
            log.info(i18n.XEH_CLEANUP)
            xeH._cleanup()
            return

    try:
        if opt.urls:
            for u in opt.urls:
                xeH.add_task(u.strip())
            # Thread(target = lambda:(time.sleep(0.618), setattr(xeH, "_exit", XEH_STATE_SOFT_EXIT))).start()
        Thread(target = xeH._task_loop, name = "main" ).start()
        while xeH._exit < XEH_STATE_CLEAN:
            # if specify urls, finished this task and exit xeHentai
            if opt.urls and not [k for k, v in xeH._all_tasks.items() if TASK_STATE_WAITING <= v.state < TASK_STATE_FINISHED]:
                xeH._exit = XEH_STATE_SOFT_EXIT
            time.sleep(1)
    except KeyboardInterrupt:
        log.info(i18n.XEH_CLEANUP)
        xeH._term_threads()
    except Exception as ex:
        log.error(i18n.XEH_CRITICAL_ERROR % traceback.format_exc())
        xeH._term_threads()
    else:
        sys.exit(0) # this is mandatory for single task auto exit
    try:
        # we should call cleanup ourself because we break out of task_loop
        xeH._cleanup()
    except KeyboardInterrupt:
        pass
    # this is mandatory for ctrl+c kill
    os._exit(0)

''' -ro --redirect-norm   是否应用在线代理到已解析到的非原图，默认不启用
    -f  --force           即使超出配额也下载，默认为否
    -j  --no-jp-name      是否不使用日语命名，默认为否'''

def _parse_range(s):
    rg = []
    s = s.replace("，", ",")
    for r in s.split(','):
        r = r.strip()
        m = re.match(r'(\d+)(?:-(\d+))?$', r)
        if not m:
            raise argparse.ArgumentTypeError(logger.safestr(i18n.c(ERR_NOT_RANGE_FORMAT) % r))
        start = int(m.group(1))
        end = int(m.group(2) or start)
        rg.append((start, end))
    return sorted(rg)

class _AddToListAction(argparse.Action):
    ''' This action add a value 'add_value' to the list 'dest' '''
    def __init__(self, option_strings, dest, add_value=None, current=None, nargs=None, **kwargs):
        super(_AddToListAction, self).__init__(option_strings, dest, default=None, nargs=0, **kwargs)
        self.add_value = add_value
        # to use in formatting output
        self.current = current
    
    def __call__(self, parser, namespace, values, option_string=None):
        if getattr(namespace, self.dest, None) is None:
            setattr(namespace, self.dest, [])
        items = list(getattr(namespace, self.dest))
        items.append(self.add_value)
        setattr(namespace, self.dest, items)

def parse_opt():
    _def = {k:v for k,v in default_config.__dict__.items() if not k.startswith("_")}
    _def.update({k:v for k,v in config.__dict__.items() if not k.startswith("_")})
    if not PY3K:
        for k in ('dir', 'log_path'):
            _def[k] = _def[k].decode('utf-8')
    parser = argparse.ArgumentParser(description = i18n.XEH_OPT_DESC, epilog = i18n.XEH_OPT_EPILOG, add_help = False)
    # the followings are handled in cli
    parser.add_argument('-u', '--username', help = i18n.XEH_OPT_u)
    parser.add_argument('-k', '--key', help = i18n.XEH_OPT_k)
    parser.add_argument('-c', '--cookie', help = i18n.XEH_OPT_c)
    parser.add_argument('-i', '--interactive', action = 'store_true', default = False,
                        help = i18n.XEH_OPT_i)
    parser.add_argument('--daemon', action = 'store_true', default = _def['daemon'],
                        help = i18n.XEH_OPT_daemon)
    # the followings are passed to xeHentai
    parser.add_argument('urls', metavar = 'url', type = str, nargs = '*',
                        help = i18n.XEH_OPT_URLS)
    # parser.add_argument('-f', '--fast-scan', action = 'store_true', default = _def.fast_scan,
    #                     help = i18n.XEH_OPT_f)
    parser.add_argument('-d', '--dir', default = os.path.abspath(_def['dir']),
                        help = i18n.XEH_OPT_d)
    parser.add_argument('-o', '--download-ori',
                        action = 'store_true', default = _def['download_ori'],
                        help = i18n.XEH_OPT_o)
    parser.add_argument('-j', '--jpn-title', type = bool, metavar = "BOOL", default = _def['jpn_title'],
                        dest = 'jpn_title', help = i18n.XEH_OPT_j)
    parser.add_argument('-r', '--rename-ori', type = bool, metavar = "BOOL", default = _def['rename_ori'],
                        help = i18n.XEH_OPT_r)

    parser.add_argument('-p', '--proxy', action = 'append', default = _def['proxy'],
                        help = i18n.XEH_OPT_p)
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--proxy-image', action = 'store_true', default = _def['proxy_image'],
                        help = i18n.XEH_OPT_proxy_image)
    group.add_argument('--proxy-image-only', action = 'store_true', default = _def['proxy_image_only'],
                        help = i18n.XEH_OPT_proxy_image_only)
    parser.add_argument('--rpc-interface', metavar = "ADDR", default = _def['rpc_interface'],
                        help = i18n.XEH_OPT_rpc_interface)
    parser.add_argument('--rpc-port', type = int, metavar = "PORT", default = _def['rpc_port'],
                        help = i18n.XEH_OPT_rpc_port)
    parser.add_argument('--rpc-secret', metavar = "...", default = _def['rpc_secret'],
                        help = i18n.XEH_OPT_rpc_secret)
    parser.add_argument('--delete-task-files', type = bool, metavar = "BOOL", default = _def['delete_task_files'],
                        dest = 'delete_task_files', help = i18n.XEH_OPT_delete_task_files)
    parser.add_argument('-a', '--archive', type = bool, metavar = "BOOL", default = _def['make_archive'],
                        dest = 'make_archive', help = i18n.XEH_OPT_a)
    parser.add_argument('--download-range', type = _parse_range, metavar = "a-b,c-d,e", default = None,
                        dest = 'download_range', help = i18n.XEH_OPT_download_range)
    parser.add_argument('-t', '--thread', type = int, metavar = 'N',
                        default = _def['download_thread_cnt'], dest = 'download_thread_cnt',
                        help = i18n.XEH_OPT_t)
    parser.add_argument('--timeout', type = int, metavar = "N", default = _def['download_timeout'],
                        dest = 'download_timeout', help = i18n.XEH_OPT_timeout)
    parser.add_argument('-f', '--force', action = _AddToListAction,
                        current = ERR_QUOTA_EXCEEDED in _def['ignored_errors'],
                        add_value = ERR_QUOTA_EXCEEDED, dest='ignored_errors',
                        help = i18n.XEH_OPT_f)

    parser.add_argument('-l', '--logpath', metavar = '/path/to/eh.log',
                        default = os.path.abspath(_def['log_path']), help = i18n.XEH_OPT_l)

    parser.add_argument('-v', '--verbose', action = 'count', default = _def['log_verbose'],
                        help = i18n.XEH_OPT_v)
    parser.add_argument('-h','--help', action = 'help', help = i18n.XEH_OPT_h)
    parser.add_argument('--version', action = 'version',
                        version = '%s v%.3f%s' % (SCRIPT_NAME, __version__, '-dev' if DEVELOPMENT else ""),
                        help = i18n.XEH_OPT_version)
    args = parser.parse_args()

    return args

def interactive(xeH):
    def _readline(x, default = ""):
        if default:
            x = x % default
        _ = input(logger.safestr(x)) if PY3K else raw_input(logger.safestr(x))
        _ = _ or default
        return _ if PY3K else _.decode(locale.getdefaultlocale()[1] or 'utf-8')

    if not xeH.has_login and _readline(i18n.PS_LOGIN) == 'y':
        uname = pwd = ""
        while not uname:
            uname = _readline(i18n.PS_USERNAME)
        while not pwd:
            pwd = _readline(i18n.PS_PASSWD)
        xeH.login_exhentai(uname, pwd)
    url = proxy = download_range = ""
    while not url:
        url = _readline(i18n.PS_URL)
    url = url.split(",")
    download_ori = _readline(i18n.PS_DOWNLOAD_ORI, 'y' if xeH.cfg['download_ori'] else 'n') == 'y'
    proxy = _readline(i18n.PS_PROXY).strip()
    proxy = [proxy] if proxy else xeH.cfg['proxy']
    __def_dir = os.path.abspath(xeH.cfg['dir'])
    # if not PY3K:
    #    __def_dir = __def_dir.decode(sys.getfilesystemencoding())
    _dir = _readline(i18n.PS_DOWNLOAD_DIR % __def_dir) or xeH.cfg['dir']
    rename_ori = _readline(i18n.PS_RENAME_ORI, 'y' if xeH.cfg['rename_ori'] else 'n') == 'y'
    make_archive = _readline(i18n.PS_MAKE_ARCHIVE, 'y' if xeH.cfg['make_archive'] else 'n') == 'y'
    jpn_title = _readline(i18n.PS_JPN_TITLE, 'y' if xeH.cfg['jpn_title'] else 'n') == 'y'
    while not download_range:
        _ = _readline(i18n.PS_DOWNLOAD_RANGE)
        if not _:
            download_range = []
            break
        try:
            download_range = _parse_range(logger.safestr(_))
        except argparse.ArgumentTypeError as ex:
            print(ex)
        else:
            break
    return {'urls': url, 'proxy': proxy, 'download_ori': download_ori, 'dir': _dir, 'rename_ori':rename_ori,
            'make_archive': make_archive, 'jpn_title': jpn_title, 'save_tasks': False,
            'download_range': download_range}



#!/usr/bin/env python
# coding:utf-8
# Contributor:
#      fffonion        <fffonion@gmail.com>

import os
import re
import sys
import uuid
import random

from ..const import *

if os.name == 'nt':
    filename_filter = re.compile("[|:?\\/*'\"<>]|\.+(?:$)")
else:# assume posix
    filename_filter = re.compile("[\/:]")

if PY3K:
    unichr = chr

def parse_cookie(coostr):
    ret = {}
    for coo in coostr.split(";"):
        coo = coo.strip()
        if coo.lower() in ('secure', 'httponly'):
            continue
        _ = coo.split("=")
        k = _[0]
        v = "=".join(_[1:])
        if k.lower() in ('path', 'expires', 'domain', 'max-age', 'comment'):
            continue
        ret[k] = v
    return ret

def make_cookie(coodict):
    return ";".join(map("=".join, coodict.items()))

def make_ua():
    rrange = lambda a, b, c = 1: c == 1 and random.randrange(a, b) or int(1.0 * random.randrange(a * c, b * c) / c)
    ua = 'Mozilla/%d.0 (Windows NT %d.%d) AppleWebKit/%d (KHTML, like Gecko) Chrome/%d.%d Safari/%d' % (
        rrange(4, 7, 10), rrange(5, 7), rrange(0, 3), rrange(535, 538, 10),
        rrange(21, 27, 10), rrange(0, 9999, 10), rrange(535, 538, 10)
    )

def get_proxy_policy(cfg):
    if cfg['proxy_image_only']:
        return RE_URL_IMAGE
    if cfg['proxy_image']:
        return RE_URL_ALL
    return RE_URL_WEBPAGE

def parse_human_time(s):
    rt = 0
    day = re.findall('(\d+)\sdays*', s)
    if day:
        rt += 86400 * int(day[0])
    hour = re.findall('(\d+)\shours*', s)
    if hour:
        rt += 3600 * int(hour[0])
    minute = re.findall('(\d+)\sminutes*', s)
    if minute:
        rt += 60 * int(minute[0])
    else:
        rt += 60
    return rt

def htmlescape(s):
    def replc(match):
        #print match.group(0),match.group(1),match.group(2)
        dict={'amp':'&','nbsp':' ','quot':'"','lt':'<','gt':'>','copy':'©','reg':'®'}
        #dict+={'∀':'forall','∂':'part','∃':'exist','∅':'empty','∇':'nabla','∈':'isin','∉':'notin','∋':'ni','∏':'prod','∑':'sum','−':'minus','∗':'lowast','√':'radic','∝':'prop','∞':'infin','∠':'ang','∧':'and','∨':'or','∩':'cap','∪':'cup','∫':'int','∴':'there4','∼':'sim','≅':'cong','≈':'asymp','≠':'ne','≡':'equiv','≤':'le','≥':'ge','⊂':'sub','⊃':'sup','⊄':'nsub','⊆':'sube','⊇':'supe','⊕':'oplus','⊗':'otimes','⊥':'perp','⋅':'sdot','Α':'Alpha','Β':'Beta','Γ':'Gamma','Δ':'Delta','Ε':'Epsilon','Ζ':'Zeta','Η':'Eta','Θ':'Theta','Ι':'Iota','Κ':'Kappa','Λ':'Lambda','Μ':'Mu','Ν':'Nu','Ξ':'Xi','Ο':'Omicron','Π':'Pi','Ρ':'Rho','Σ':'Sigma','Τ':'Tau','Υ':'Upsilon','Φ':'Phi','Χ':'Chi','Ψ':'Psi','Ω':'Omega','α':'alpha','β':'beta','γ':'gamma','δ':'delta','ε':'epsilon','ζ':'zeta','η':'eta','θ':'theta','ι':'iota','κ':'kappa','λ':'lambda','μ':'mu','ν':'nu','ξ':'xi','ο':'omicron','π':'pi','ρ':'rho','ς':'sigmaf','σ':'sigma','τ':'tau','υ':'upsilon','φ':'phi','χ':'chi','ψ':'psi','ω':'omega','ϑ':'thetasym','ϒ':'upsih','ϖ':'piv','Œ':'OElig','œ':'oelig','Š':'Scaron','š':'scaron','Ÿ':'Yuml','ƒ':'fnof','ˆ':'circ','˜':'tilde',' ':'ensp',' ':'emsp',' ':'thinsp','‌':'zwnj','‍':'zwj','‎':'lrm','‏':'rlm','–':'ndash','—':'mdash','‘':'lsquo','’':'rsquo','‚':'sbquo','“':'ldquo','”':'rdquo','„':'bdquo','†':'dagger','‡':'Dagger','•':'bull','…':'hellip','‰':'permil','′':'prime','″':'Prime','‹':'lsaquo','›':'rsaquo','‾':'oline','€':'euro','™':'trade','←':'larr','↑':'uarr','→':'rarr','↓':'darr','↔':'harr','↵':'crarr','⌈':'lceil','⌉':'rceil','⌊':'lfloor','⌋':'rfloor','◊':'loz','♠':'spades','♣':'clubs','♥':'hearts','♦':'diams'}
        if len(match.groups()) >= 2:
            if match.group(1) == '#':
                return unichr(int(match.group(2)))
            else:
                return dict.get(match.group(2), '?')
    htmlre = re.compile("&(#?)(\d{1,5}|\w{1,8}|[a-z]+);")
    return htmlre.sub(replc, s)

def legalpath(s):
    sanitized = filename_filter.sub(lambda x:"", s)
    # windows doesn't like trailing while spaces
    if os.name == 'nt':
        sanitized = sanitized.rstrip()
    return sanitized



# coding:utf-8

from ..const import *

err_msg = {
    ERR_URL_NOT_RECOGNIZED: "网址不够绅士",
    ERR_CANT_DOWNLOAD_EXH: "需要登录后才能下载里站",
    ERR_ONLY_VISIBLE_EXH: "这个本子只有里站能看到",
    ERR_MALFORMED_HATHDL: "hathdl文件有猫饼，解析失败",
    ERR_GALLERY_REMOVED: "这个本子被移除了，大概里站能看到",
    ERR_KEY_EXPIRED: "下载链接不太正常",
    ERR_NO_PAGEURL_FOUND: "没有找到页面链接，网站改版了嘛？",
    ERR_CONNECTION_ERROR: "连接有问题？",
    ERR_IP_BANNED: "IP被ban了, 恢复时间: %s",
    ERR_IMAGE_BROKEN: "下载的图片有猫饼",
    ERR_QUOTA_EXCEEDED: "配额超限",
    ERR_TASK_NOT_FOUND: "没有该GUID对应的任务",
    ERR_TASK_LEVEL_UNDEF: "任务过滤等级不存在",
    ERR_DELETE_RUNNING_TASK: "无法删除运行中的任务",
    ERR_TASK_CANNOT_PAUSE: "这个任务无法被暂停",
    ERR_TASK_CANNOT_RESUME: "这个任务无法被恢复",
    ERR_CANNOT_CREATE_DIR: "无法创建文件夹 %s",
    ERR_CANNOT_MAKE_ARCHIVE: "无法制作压缩包 %s",
    ERR_NOT_RANGE_FORMAT: "'%s'不符合范围的格式, 正确的格式为 1-3 或者 5",
#    ERR_HATHDL_NOTFOUND: "hathdl文件未找到"
    ERR_RPC_PARSE_ERROR: "Parse error.",
    ERR_RPC_INVALID_REQUEST: "Invalid request.",
    ERR_RPC_METHOD_NOT_FOUND: "Method not found.",
    ERR_RPC_INVALID_PARAMS: "Invalid method parameter(s).",
    ERR_RPC_UNAUTHORIZED: "Unauthorized",
    ERR_RPC_EXEC_ERROR: "",
    ERR_SAVE_SESSION_FAILED: "",
}

ERR_NOMSG = "未指定的错误，错误号 %d"

XEH_OPT_DESC = "绅♂士下载器"
XEH_OPT_EPILOG = "如果参数未指定，则使用config.py中的默认值; " \
        "讨论和反馈问题：https://yooooo.us/2013/xehentai"
XEH_OPT_URLS = "下载页的网址"
XEH_OPT_u = "用户名"
XEH_OPT_k = "密码"
XEH_OPT_c = "Cookie字符串，如果指定了用户名和密码，此项会被忽略"
XEH_OPT_o = "是否下载原始图片（如果存在），需要登录 (当前: %(default)s)"
XEH_OPT_t = "下载线程数 (当前: %(default)d)"
# XEH_OPT_f = "快速扫描，从hathdl猜测页面链接，但有时会抽风 (当前: %(default)s)"
XEH_OPT_l = "保存日志的路径 (当前: %(default)s)"
XEH_OPT_p = "设置代理, 可以指定多次, 当前支持的类型: socks5/4a, http(s), glype. 代理默认只用于扫描网页 (当前: %(default)s)"
XEH_OPT_proxy_image = "同时使用代理来下载图片和扫描网页（当前: %(default)s)"
XEH_OPT_proxy_image_only = "仅使用代理来下载图片, 不用于扫描网页 (当前: %(default)s)"
XEH_OPT_d = "设置下载目录 (当前: %(default)s)"
XEH_OPT_v = "设置日志装逼等级 (当前: %(default)s)"
XEH_OPT_i = "交互模式，如果开启后台模式，此项会被忽略 (当前: %(default)s)"
XEH_OPT_r = "将图片重命名为原始名称，如果关闭则使用序号 (当前: %(default)s)"
XEH_OPT_daemon = "后台模式 (当前: %(default)s)"
XEH_OPT_rpc_interface = "设置JSON-RPC监听IP (当前: %(default)s)"
XEH_OPT_rpc_port = "设置JSON-RPC监听端口 (当前: %(default)s)"
XEH_OPT_rpc_secret = "设置JSON-RPC密钥 (当前: %(default)s)"
XEH_OPT_a = "下载完成后生成zip压缩包并删除下载目录 (当前: %(default)s)"
XEH_OPT_delete_task_files = "删除任务时同时删除下载的文件 (current: %(default)s)"
XEH_OPT_j = "使用日语标题, 如果关闭则使用英文或罗马字标题 (当前: %(default)s)"
XEH_OPT_download_range = "设置下载的图片范围, 格式为 开始位置-结束位置, 或者单张图片的位置, " \
"使用逗号来分隔多个范围, 例如 5-10,15,20-25, 默认为下载所有"
XEH_OPT_timeout = "设置下载图片的超时 (当前: %(default)s秒)"
XEH_OPT_f = "忽略配额判断，继续下载 (当前: %(current)s)"
XEH_OPT_h = "显示本帮助信息"
XEH_OPT_version = "显示版本信息"
XEH_OPT_IGNORING_I = "后台模式已忽略 -i 参数"


PS_LOGIN = "当前没有登陆，要登陆吗 (y/n)? > "
PS_USERNAME = "输入用户名 > "
PS_PASSWD = "输入密码   > "
PS_URL = "输入地址（使用,分割下载多个）> "
PS_PROXY = "输入代理地址 (可选) > "
PS_DOWNLOAD_ORI = "是否下载原图（默认%s） (y/n)? > "
PS_RENAME_ORI  = "是否自动重命名（默认%s） (y/n)? > "
PS_MAKE_ARCHIVE = "是否制作zip压缩包（默认%s） (y/n)? > "
PS_JPN_TITLE = "是否使用日语标题（默认%s） (y/n)? > "
PS_DOWNLOAD_RANGE = "下载范围, 使用逗号分割多个范围, 回车下载全部 > "
PS_DOWNLOAD_DIR = "下载目录 (当前: %s)\n回车确认或输入新路径 > "

PROXY_CANDIDATE_CNT = "代理池中有%d个代理"

TASK_PUT_INTO_WAIT = "任务 #%s 已存在, 加入等待队列"
TASK_ERROR = "任务 #%s 发生错误: %s"
TASK_MIGRATE_EXH = "任务 #%s 使用里站地址重新下载"
TASK_TITLE = "任务 #%s 标题 %s"
TASK_WILL_DOWNLOAD_CNT = "任务 #%s 将下载%d个文件，共%d个 "
TASK_START = "任务 #%s 开始"
TASK_FINISHED = "任务 #%s 下载完成"
TASK_START_PAGE_RESCAN = "任务 #%s 图片被缩放，进行完整扫描"
# TASK_FAST_SCAN = "任务 #%s 使用快速扫描"
TASK_START_MAKE_ARCHIVE = "任务 #%s 开始打包"
TASK_MAKE_ARCHIVE_FINISHED = "任务 #%s 打包完成，保存在: %s, 用时%.1f秒"
TASK_STOP_QUOTA_EXCEEDED = "任务 #%s 配额超限"
TASK_STUCK = "任务 #%s 卡住了, 可能是脚本有bug, 或者网络连接太慢了"

XEH_STARTED = "xeHentai %s 已启动"
XEH_LOOP_FINISHED = "程序循环已完成"
XEH_LOGIN_EXHENTAI = "登录绅士"
XEH_LOGIN_OK = "已成为绅士"
XEH_LOGIN_FAILED = "无法登录绅士；检查输入是否有误或者换一个帐号。\n推荐在浏览器登录后使用RPC复制cookie到xeHentai (教程: http://t.cn/Rctr4Pf)"
XEH_LOAD_TASKS_CNT = "从存档中读取了%d个任务"
XEH_LOAD_OLD_COOKIE = "从1.x版cookie文件从读取了登录信息"
XEH_DAEMON_START = "后台进程已启动，PID为%d"
XEH_PLATFORM_NO_DAEMON = "后台模式不支持您的系统: %s"
XEH_CLEANUP = "擦干净..."
XEH_CRITICAL_ERROR = "xeHentai 抽风啦:\n%s"
XEH_DOWNLOAD_ORI_NEED_LOGIN = "下载原图需要登录"
XEH_FILE_DOWNLOADED = "绅士-{} 已下载图片 #{} {}"
XEH_RENAME_HAS_ERRORS = "部分图片重命名失败:\n%s"
XEH_DOWNLOAD_HAS_ERROR = "绅士-%s 下载图片时出错: %s, 将在稍后重试"

RPC_STARTED = "RPC服务器监听在 %s:%d"
RPC_TOO_OPEN = "RPC服务器监听在公网IP (%s)，为了安全起见应该设置rpc_secret"
RPC_CANNOT_BIND = "RPC服务器无法启动：%s"

SESSION_LOAD_EXCEPTION = "读取存档时遇到错误: %s"
SESSION_WRITE_EXCEPTION = "写入存档时遇到错误: %s"

THREAD = "绅士"
THREAD_UNCAUGHT_EXCEPTION = "绅士-%s 未捕获的异常\n%s"
THREAD_MAY_BECOME_ZOMBIE = "绅士-%s 可能变成了丧尸"
THREAD_SWEEP_OUT = "绅士-%s 挂了, 不再理它"

QUEUE = "队列"

PROXY_DISABLE_BANNED = "禁用了一个被ban的代理，将在约%s秒后恢复"



#!/usr/bin/env python

PROJ_NAME = 'xeHentai'
PACKAGE_NAME = 'xeHentai'

PROJ_METADATA = '%s.json' % PROJ_NAME

import os, json, imp
here = os.path.abspath(os.path.dirname(__file__))

try:
    README = open(os.path.join(here, 'README.md')).read()
except:
    README = ""
try:
    CHANGELOG = open(os.path.join(here, 'CHANGELOG.md')).read()
except:
    CHANGELOG = ""
VERSION = imp.load_source('version', os.path.join(here, '%s/const.py' % PACKAGE_NAME)).__version__

packages = [
    'xeHentai',
    'xeHentai.util',
    'xeHentai.i18n',
]
requires = ['requests']

from setuptools import setup

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description='xeHentai Downloader',
    long_description=README + '\n\n' + CHANGELOG,
    long_description_content_type="text/markdown",
    author='fffonion',
    author_email='fffonion@gmail.com',
    url='https://yooooo.us/2013/xehentai',
    packages=packages,
    package_dir={'requests': 'requests'},
    include_package_data=True,
    install_requires=requires,
    license='GPLv3',
    zip_safe=False,
    classifiers=(
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: End Users/Desktop',
        'Natural Language :: English',
        'Environment :: Console',
        'Environment :: MacOS X',
        'Environment :: Win32 (MS Windows)',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: Chinese (Simplified)',
        'Natural Language :: Chinese (Traditional)',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ),
    requires=requires,
    entry_points = {'console_scripts': ["xeH = xeHentai.cli:start"]},
)



#!/usr/bin/env python
# coding:utf-8
# Contributor:
#      fffonion        <fffonion@gmail.com>

import importlib
from ..const import *
from . import en_us as lng_fallback

try:
    _locale = LOCALE.lower() if LOCALE else 'en_us'
    if _locale in ('zh_cn', 'zh_sg'):
        _locale = 'zh_hans'
    elif _locale in ('zh_tw', 'zh_hk', 'zh_mo'):
        _locale = 'zh_hant'
    lng = importlib.import_module("%s.i18n.%s" % (SCRIPT_NAME, _locale))
except (ImportError, ValueError):
    lng = lng_fallback


class _(object):
    def c(cls, code):
        _ = code not in lng.err_msg and \
            (code not in lng_fallback.err_msg and \
                (cls.ERR_NOMSG % code) or \
                    lng_fallback.err_msg[code] ) or \
            lng.err_msg[code]
        return _ if PY3K else (
            _ if isinstance(_, unicode) else _.decode('utf-8')) # cls.ERR_NOMSG % code is unicode

    def __getattr__(cls, idx):
        _ = not hasattr(lng, idx) and \
            getattr(lng_fallback, idx) or \
            getattr(lng, idx)
        return _ if PY3K else _.decode('utf-8')

i18n = _()
